"""
Simple implementation of International Standard Atmosphere.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numbers import Number
from typing import Sequence, Union

import numpy as np
from deprecated import deprecated
from scipy.constants import R, atmosphere, foot

AIR_MOLAR_MASS = 28.9647e-3
AIR_GAS_CONSTANT = R / AIR_MOLAR_MASS
SEA_LEVEL_PRESSURE = atmosphere
SEA_LEVEL_TEMPERATURE = 288.15
TROPOPAUSE = 11000


@deprecated(
    version="1.2.0", reason="Will be removed in version 2.0. Please use stdatm.Atmosphere instead"
)
class Atmosphere:
    """
    Simple implementation of International Standard Atmosphere
    for troposphere and stratosphere.

    Atmosphere properties are provided in the same "shape" as provided
    altitude:

    - if altitude is given as a float, returned values will be floats
    - if altitude is given as a sequence (list, 1D numpy array, ...), returned
      values will be 1D numpy arrays
    - if altitude is given as nD numpy array, returned values will be nD numpy
      arrays

    Usage:

    .. code-block::

        >>> pressure = Atmosphere(30000).pressure # pressure at 30,000 feet, dISA = 0 K
        >>> density = Atmosphere(5000, 10).density # density at 5,000 feet, dISA = 10 K


        >>> atm = Atmosphere(np.arange(0,10001,1000, 15)) # init for alt. 0 to 10,000, dISA = 15K
        >>> temperatures = atm.pressure # pressures for all defined altitudes
        >>> viscosities = atm.kinematic_viscosity # viscosities for all defined altitudes
    """

    # pylint: disable=too-many-instance-attributes  # Needed for avoiding redoing computations
    def __init__(
        self,
        altitude: Union[float, Sequence[float]],
        delta_t: float = 0.0,
        altitude_in_feet: bool = True,
    ):
        """
        :param altitude: altitude (units decided by altitude_in_feet)
        :param delta_t: temperature increment (°C) applied to whole temperature profile
        :param altitude_in_feet: if True, altitude should be provided in feet. Otherwise,
                                 it should be provided in meters.
        """
        self.delta_t = delta_t

        # Floats will be provided as output if altitude is a scalar
        self._float_expected = isinstance(altitude, Number)

        # For convenience, let's have altitude as numpy arrays and in meters in all cases
        unit_coeff = foot if altitude_in_feet else 1.0
        self._altitude = np.asarray(altitude) * unit_coeff

        # Sets indices for tropopause
        self._idx_tropo = self._altitude < TROPOPAUSE
        self._idx_strato = self._altitude >= TROPOPAUSE

        # Outputs
        self._temperature = None
        self._pressure = None
        self._density = None
        self._speed_of_sound = None
        self._kinematic_viscosity = None
        self._mach = None
        self._equivalent_airspeed = None
        self._true_airspeed = None
        self._unitary_reynolds = None

    def get_altitude(self, altitude_in_feet: bool = True) -> Union[float, Sequence[float]]:
        """
        :param altitude_in_feet: if True, altitude is returned in feet. Otherwise,
                                 it is returned in meters
        :return: altitude provided at instantiation
        """
        if altitude_in_feet:
            return self._return_value(self._altitude / foot)
        return self._return_value(self._altitude)

    @property
    def delta_t(self) -> Union[float, Sequence[float]]:
        """Temperature increment applied to whole temperature profile."""
        return self._delta_t

    @delta_t.setter
    def delta_t(self, value: Union[float, Sequence[float]]):
        self._delta_t = np.asarray(value)

    @property
    def temperature(self) -> Union[float, Sequence[float]]:
        """Temperature in K."""
        if self._temperature is None:
            self._temperature = np.zeros(self._altitude.shape)
            self._temperature[self._idx_tropo] = (
                SEA_LEVEL_TEMPERATURE - 0.0065 * self._altitude[self._idx_tropo] + self._delta_t
            )
            self._temperature[self._idx_strato] = 216.65 + self._delta_t
        return self._return_value(self._temperature)

    @property
    def pressure(self) -> Union[float, Sequence[float]]:
        """Pressure in Pa."""
        if self._pressure is None:
            self._pressure = np.zeros(self._altitude.shape)
            self._pressure[self._idx_tropo] = (
                SEA_LEVEL_PRESSURE
                * (1 - (self._altitude[self._idx_tropo] / 44330.78)) ** 5.25587611
            )
            self._pressure[self._idx_strato] = 22632 * 2.718281 ** (
                1.7345725 - 0.0001576883 * self._altitude[self._idx_strato]
            )
        return self._return_value(self._pressure)

    @property
    def density(self) -> Union[float, Sequence[float]]:
        """Density in kg/m3."""
        if self._density is None:
            self._density = self.pressure / AIR_GAS_CONSTANT / self.temperature
        return self._return_value(self._density)

    @property
    def speed_of_sound(self) -> Union[float, Sequence[float]]:
        """Speed of sound in m/s."""
        if self._speed_of_sound is None:
            self._speed_of_sound = (1.4 * AIR_GAS_CONSTANT * self.temperature) ** 0.5
        return self._return_value(self._speed_of_sound)

    @property
    def kinematic_viscosity(self) -> Union[float, Sequence[float]]:
        """Kinematic viscosity in m2/s."""
        if self._kinematic_viscosity is None:
            self._kinematic_viscosity = (
                (0.000017894 * (self.temperature / SEA_LEVEL_TEMPERATURE) ** (3 / 2))
                * ((SEA_LEVEL_TEMPERATURE + 110.4) / (self.temperature + 110.4))
            ) / self.density
        return self._return_value(self._kinematic_viscosity)

    @property
    def mach(self) -> Union[float, Sequence[float]]:
        """Mach number."""
        if self._mach is None and self.true_airspeed is not None:
            self._mach = self.true_airspeed / self.speed_of_sound
        return self._return_value(self._mach)

    @property
    def true_airspeed(self) -> Union[float, Sequence[float]]:
        """True airspeed (TAS) in m/s."""
        # Dev note: true_airspeed is the "hub". Other speed values will be calculated
        # from this true_airspeed.
        if self._true_airspeed is None:
            if self._mach is not None:
                self._true_airspeed = self._mach * self.speed_of_sound
            if self._equivalent_airspeed is not None:
                sea_level = Atmosphere(0)
                self._true_airspeed = self._equivalent_airspeed * np.sqrt(
                    sea_level.density / self.density
                )
            if self._unitary_reynolds is not None:
                self._true_airspeed = self._unitary_reynolds * self.kinematic_viscosity
        return self._return_value(self._true_airspeed)

    @property
    def equivalent_airspeed(self) -> Union[float, Sequence[float]]:
        """Equivalent airspeed (EAS) in m/s."""
        if self._equivalent_airspeed is None and self.true_airspeed is not None:
            sea_level = Atmosphere(0)
            self._equivalent_airspeed = self.true_airspeed / np.sqrt(
                sea_level.density / self.density
            )

        return self._return_value(self._equivalent_airspeed)

    @property
    def unitary_reynolds(self) -> Union[float, Sequence[float]]:
        """Unitary Reynolds number in 1/m."""
        if self._unitary_reynolds is None and self.true_airspeed is not None:
            self._unitary_reynolds = self.true_airspeed / self.kinematic_viscosity
        return self._return_value(self._unitary_reynolds)

    @mach.setter
    def mach(self, value: Union[float, Sequence[float]]):
        self._reset_speeds()
        self._mach = np.asarray(value)

    @true_airspeed.setter
    def true_airspeed(self, value: Union[float, Sequence[float]]):
        self._reset_speeds()
        self._true_airspeed = np.asarray(value)

    @equivalent_airspeed.setter
    def equivalent_airspeed(self, value: Union[float, Sequence[float]]):
        self._reset_speeds()
        self._equivalent_airspeed = np.asarray(value)

    @unitary_reynolds.setter
    def unitary_reynolds(self, value: Union[float, Sequence[float]]):
        self._reset_speeds()
        self._unitary_reynolds = np.asarray(value)

    def _reset_speeds(self):
        """To be used before setting a new speed value as private attribute."""
        self._mach = None
        self._true_airspeed = None
        self._equivalent_airspeed = None
        self._unitary_reynolds = None

    def _return_value(self, value):
        """
        :returns: a float when needed. Otherwise, returns the value itself.
        """
        if self._float_expected and value is not None:
            try:
                # It's faster to try... catch than to test np.size(value).
                # (but float(value) is slow to fail if value is None, so
                #  it is why we test it before)
                return float(value)
            except TypeError:
                pass
        return value


@deprecated(
    version="1.2.0", reason="Will be removed in version 2.0. Please use stdatm.AtmosphereSI instead"
)
class AtmosphereSI(Atmosphere):
    """Same as :class:`Atmosphere` except that altitudes are always in meters."""

    def __init__(self, altitude: Union[float, Sequence[float]], delta_t: float = 0.0):
        """
        :param altitude: altitude in meters
        :param delta_t: temperature increment (°C) applied to whole temperature profile
        """
        super().__init__(altitude, delta_t, altitude_in_feet=False)

    @property
    def altitude(self):
        """Altitude in meters."""
        return self.get_altitude(altitude_in_feet=False)



import numpy as np

import openmdao.api as om


class ComputeOwe(om.ExplicitComponent):
    """
    Computes the aircraft structural mass based on its MTOW and wing mass
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="mtow", units="kg", val=np.nan)
        self.add_input(name="wing_mass", units="kg", val=np.nan)

        # Defining the output(s)

        self.add_output(name="owe", units="kg")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        mtow = inputs["mtow"]
        wing_mass = inputs["wing_mass"]

        # Let's start by computing the weight of the aircraft without the wings
        owe_without_wing = mtow * (0.43 + 0.0066 * np.log(mtow))

        # Let's now add the wing mass to get the structural weight
        owe = owe_without_wing + wing_mass

        outputs["owe"] = owe



"""
Exception for cmd package
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from fastoad.exceptions import FastError


class FastPathExistsError(FastError):
    """Raised when asked for writing a file/folder that already exists."""

    def __init__(self, *args):
        super().__init__(*args)
        self.file_path = args[1]


class FastNoAvailableNotebookError(FastError):
    """Raised when no notebook is available for creation."""

    def __init__(self, distribution_name=None):
        msg = "No notebook available "
        if distribution_name:
            msg += f'in installed package "{distribution_name}".'
        else:
            msg += "in FAST-OAD plugins."

        self.distribution_name = distribution_name
        super().__init__(msg)



import numpy as np

import openmdao.api as om


class ComputeProfileDrag(om.ExplicitComponent):
    """
    Computes the profile drag of the aircraft based on the wing area
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="wing_area", units="m**2", val=np.nan)

        # Defining the output(s)

        self.add_output(name="profile_drag_coefficient")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        wing_area = inputs["wing_area"]

        # Wet area of the aircraft without the wings
        wing_area_ref = 13.50

        # Profile drag coefficient of the aircraft without the wings
        cd0_other = 0.022

        # Constant linking the wing profile drag to its wet area, and by extension, its reference area
        c = 0.0004

        # Computation of the profile drag
        cd0 = cd0_other * wing_area_ref / wing_area + c

        outputs["profile_drag_coefficient"] = cd0



#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import warnings
from copy import deepcopy
from typing import Tuple

import numpy as np
import openmdao
import openmdao.api as om
from openmdao.core.constants import _SetupStatus
from openmdao.core.system import _MetadataDict
from packaging import version

from fastoad.io import DataFile, VariableIO
from fastoad.module_management.service_registry import RegisterSubmodel
from fastoad.openmdao.validity_checker import ValidityDomainChecker
from fastoad.openmdao.variables import VariableList
from ._utils import problem_without_mpi
from .exceptions import FASTOpenMDAONanInInputFile
from ..module_management._bundle_loader import BundleLoader

# Name of IVC that will contain input values
INPUT_SYSTEM_NAME = "fastoad_inputs"

# Name of IVC that will temporarily set shapes for dynamically shaped inputs
SHAPER_SYSTEM_NAME = "fastoad_shaper"


class FASTOADProblem(om.Problem):
    """
    Vanilla OpenMDAO Problem except that it can write its outputs to a file.

    It also runs :class:`~fastoad.openmdao.validity_checker.ValidityDomainChecker`
    after each :meth:`run_model` or :meth:`run_driver`
    (but it does nothing if no check has been registered).
    """

    def __init__(self, *args, **kwargs):
        if (
            version.parse(openmdao.__version__) >= version.parse("3.17")
            and "OPENMDAO_REPORTS" not in os.environ
            and "reports" not in kwargs
            and len(args) < 5
        ):
            # Automatic reports are deactivated for FAST-OAD, unless OPENMDAO_REPORTS env
            # variable is set.
            kwargs["reports"] = None
        super().__init__(*args, **kwargs)

        #: File path where :meth:`read_inputs` will read inputs
        self.input_file_path = None
        #: File path where :meth:`write_outputs` will write outputs
        self.output_file_path = None

        #: Variables that are not part of the problem but that should be written in output file.
        self.additional_variables = None

        #: If True inputs will be read after setup.
        self._read_inputs_after_setup = False

        self.model = FASTOADModel()

    def run_model(self, case_prefix=None, reset_iter_counts=True):
        status = super().run_model(case_prefix, reset_iter_counts)
        ValidityDomainChecker.check_problem_variables(self)
        BundleLoader().clean_memory()
        return status

    def run_driver(self, case_prefix=None, reset_iter_counts=True):
        status = super().run_driver(case_prefix, reset_iter_counts)
        ValidityDomainChecker.check_problem_variables(self)
        BundleLoader().clean_memory()
        return status

    def setup(self, *args, **kwargs):
        """
        Set up the problem before run.
        """
        with problem_without_mpi(self) as problem_copy:
            try:
                super(FASTOADProblem, problem_copy).setup(*args, **kwargs)
            except RuntimeError:
                vars_metadata = self._get_undetermined_dynamic_vars_metadata(problem_copy)
                if vars_metadata:
                    # If vars_metadata is empty, it means the RuntimeError was not because
                    # of dynamic shapes, and the incoming self.setup() will raise it.
                    ivc = om.IndepVarComp()
                    for name, meta in vars_metadata.items():
                        # We use a (2,)-shaped array as value here. This way, it will be easier
                        # to identify dynamic-shaped data in an input file generated from current
                        # problem.
                        ivc.add_output(name, [np.nan, np.nan], units=meta["units"])
                    self.model.add_subsystem(SHAPER_SYSTEM_NAME, ivc, promotes=["*"])

        super().setup(*args, **kwargs)

        if self._read_inputs_after_setup:
            self._read_inputs_with_setup_done()
        BundleLoader().clean_memory()

    def write_outputs(self):
        """
        Writes all outputs in the configured output file.
        """
        if self.output_file_path:
            writer = VariableIO(self.output_file_path)

            if self.additional_variables is None:
                self.additional_variables = []
            variables = VariableList(self.additional_variables)
            for var in variables:
                var.is_input = None
            variables.update(
                VariableList.from_problem(self, promoted_only=True), add_variables=True
            )
            writer.write(variables)

    def read_inputs(self):
        """
        Reads inputs of the problem.
        """
        if self._metadata and self._metadata["setup_status"] == _SetupStatus.POST_SETUP:
            self._read_inputs_with_setup_done()
        else:
            self._read_inputs_without_setup_done()
            # Input reading still needs to be done after setup, to ensure that IVC values
            # will be properly set by new inputs.
            self._read_inputs_after_setup = True

    def _get_problem_inputs(self) -> Tuple[VariableList, VariableList]:
        """
        Reads input file for the configured problem.

        Needed variables and unused variables are
        returned as a VariableList instance.

        :return: VariableList of needed input variables, VariableList with unused variables.
        """

        problem_variables = VariableList().from_problem(self)
        problem_inputs_names = [var.name for var in problem_variables if var.is_input]

        input_variables = DataFile(self.input_file_path)

        unused_variables = VariableList(
            [var for var in input_variables if var.name not in problem_inputs_names]
        )
        for name in unused_variables.names():
            del input_variables[name]

        nan_variable_names = [var.name for var in input_variables if np.all(np.isnan(var.value))]
        if nan_variable_names:
            raise FASTOpenMDAONanInInputFile(self.input_file_path, nan_variable_names)

        return input_variables, unused_variables

    def _read_inputs_with_setup_done(self):
        """
        Set initial values of inputs. self.setup() must have been run.
        """
        input_variables, unused_variables = self._get_problem_inputs()
        self.additional_variables = unused_variables
        for input_var in input_variables:
            # set_val() will crash if input_var.metadata["val"] is a list, so
            # we ensure it is a numpy array
            input_var.metadata["val"] = np.asarray(input_var.metadata["val"])

            self.set_val(input_var.name, val=input_var.val, units=input_var.units)

    def _read_inputs_without_setup_done(self):
        """
        Set initial values of inputs. self.setup() must have NOT been run.

        Input values that match an existing IVC are not taken into account
        """
        input_variables, unused_variables = self._get_problem_inputs()
        self.additional_variables = unused_variables
        with problem_without_mpi(self):
            tmp_prob = deepcopy(self)
            tmp_prob.setup()
            # At this point, there may be non-fed dynamically shaped inputs, so the setup may
            # create the "shaper" IVC, but we ignore it because we need to redefine these variables
            # in input file.
            ivc_vars = tmp_prob.model.get_io_metadata(
                "output", tags="indep_var", excludes=f"{SHAPER_SYSTEM_NAME}.*"
            )
        for meta in ivc_vars.values():
            try:
                del input_variables[meta["prom_name"]]
            except ValueError:
                pass
        if input_variables:
            self._insert_input_ivc(input_variables.to_ivc())

    def _insert_input_ivc(self, ivc: om.IndepVarComp, subsystem_name=INPUT_SYSTEM_NAME):
        with problem_without_mpi(self) as tmp_prob:
            tmp_prob.setup()

            # We get order from copied problem, but we have to ignore the "shaper"
            # and the auto IVCs.
            previous_order = [
                system.name
                for system in tmp_prob.model.system_iter(recurse=False)
                if system.name != "_auto_ivc" and system.name != SHAPER_SYSTEM_NAME
            ]

        self.model.add_subsystem(subsystem_name, ivc, promotes=["*"])
        self.model.set_order([subsystem_name] + previous_order)

    @classmethod
    def _get_undetermined_dynamic_vars_metadata(cls, problem):
        """
        Provides dict (name, metadata) for dynamically shaped inputs that are not
        fed by an existing output (assuming overall variable promotion).

        Assumes problem.setup() has been run, at least partially.

        :param problem:
        """
        # First all outputs are identified. If a dynamically shaped input is fed by a matching
        # output, its shaped will be determined.
        output_var_names = []
        for system in problem.model.system_iter(recurse=False):
            io_metadata = cls._get_io_metadata(system, "output")
            output_var_names += [meta["prom_name"] for meta in io_metadata.values()]

        dynamic_vars = {}
        for system in problem.model.system_iter(recurse=False):
            io_metadata = cls._get_io_metadata(system, "input")
            dynamic_vars.update(
                {
                    meta["prom_name"]: meta
                    for name, meta in io_metadata.items()
                    if meta["shape_by_conn"] and meta["prom_name"] not in output_var_names
                }
            )
        return dynamic_vars

    @staticmethod
    def _get_io_metadata(
        system,
        iotypes,
    ):
        # In OpenMDAO >3.16, get_io_metadata() won't complain after dynamically shaped, non-
        # connected inputs.
        if version.parse(openmdao.__version__) >= version.parse("3.17"):
            return system.get_io_metadata(iotypes)

        # For OpenMDAO<=3.16, we try the vanilla get_io_metadata() and if it fails, we
        # try with our simplified implementation.
        try:
            return system.get_io_metadata(iotypes)
        except RuntimeError:
            prefix = system.pathname + "." if system.pathname else ""
            rel_idx = len(prefix)
            if isinstance(iotypes, str):
                iotypes = (iotypes,)

            result = {}
            for iotype in iotypes:
                for abs_name, prom in system._var_abs2prom[iotype].items():
                    rel_name = abs_name[rel_idx:]
                    meta = system._var_allprocs_abs2meta[iotype].get(abs_name)
                    ret_meta = _MetadataDict(meta) if meta is not None else None
                    if ret_meta is not None:
                        ret_meta["prom_name"] = prom
                        result[rel_name] = ret_meta

            warnings.warn(
                "Dynamically shaped problem inputs are better managed with OpenMDAO>3.16 "
                "Upgrade is recommended.",
                DeprecationWarning,
            )
            return result


class AutoUnitsDefaultGroup(om.Group):
    """
    OpenMDAO group that automatically use self.set_input_defaults() to resolve declaration
    conflicts in variable units.
    """

    def configure(self):
        var_units = {}
        system: om.Group
        for system in self.system_iter(recurse=False):
            system_metadata = system.get_io_metadata("input", metadata_keys=["units"])
            var_units.update(
                {
                    metadata["prom_name"]: metadata["units"]
                    for name, metadata in system_metadata.items()
                    if "." not in metadata["prom_name"]  # tells that var is promoted
                }
            )
        for name, units in var_units.items():
            self.set_input_defaults(name, units=units)


class FASTOADModel(AutoUnitsDefaultGroup):
    """
    OpenMDAO group that defines active submodels after the initialization
    of all its subsystems, and inherits from :class:`AutoUnitsDefaultGroup` for resolving
    declaration conflicts in variable units.

    It allows to have a submodel choice in the initialize() method of a FAST-OAD module, but
    to possibly override it with the definition of :attr:`active_submodels` (i.e. from the
    configuration file).
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        #: Definition of active submodels that will be applied during setup()
        self.active_submodels = {}

    def setup(self):
        RegisterSubmodel.active_models.update(self.active_submodels)



import numpy as np

import openmdao.api as om


class UpdateMTOW(om.ExplicitComponent):
    """
    Computes new mtow based on the mission fuel and structural weight from previous iteration
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="owe", units="kg", val=np.nan)
        self.add_input(name="payload", units="kg", val=np.nan)
        self.add_input(name="mission_fuel", units="kg", val=np.nan)

        # Defining the output(s)

        self.add_output(name="mtow", units="kg")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        owe = inputs["owe"]
        payload = inputs["payload"]
        mission_fuel = inputs["mission_fuel"]

        # Let's simply add the weight we computed
        mtow_new = owe + payload + mission_fuel

        outputs["mtow"] = mtow_new



"""
Defines the analysis and plotting functions for postprocessing regarding the mission
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth
from typing import Union

import ipywidgets as widgets
import pandas as pd
import plotly.graph_objects as go
from IPython.display import clear_output, display


class MissionViewer:
    """
    A class for facilitating the post-processing of mission and trajectories
    """

    def __init__(self):
        # The dataframes containing each mission
        self.missions = {}

        # The figure displayed
        self._fig = None

        # The x selector
        self._x_widget = None

        # The y selector
        self._y_widget = None

    def add_mission(self, mission_data: Union[str, pd.DataFrame], name=None):
        """
        Adds the mission to the mission database (self.missions)
        :param mission_data: path of the mission file or Dataframe containing the mission data
        :param name: name to give to the mission
        """
        if (
            isinstance(mission_data, str)
            and mission_data.endswith(".csv")
            and pth.exists(mission_data)
        ):
            self.missions[name] = pd.read_csv(mission_data, index_col=0)
        elif isinstance(mission_data, pd.DataFrame):
            self.missions[name] = mission_data
        else:
            raise TypeError("Unknown type for mission data, please use .csv of DataFrame")

        # Initialize widgets when first mission is added
        if len(self.missions) == 1:
            self._initialize_widgets()

    def _initialize_widgets(self):
        """
        Initializes the widgets for selecting x and y
        """

        key = list(self.missions)[0]
        keys = self.missions[key].keys()

        # By default ground distance
        self._x_widget = widgets.Dropdown(value=keys[2], options=keys)
        self._x_widget.observe(self.display, "value")
        # By default altitude
        self._y_widget = widgets.Dropdown(value=keys[1], options=keys)
        self._y_widget.observe(self.display, "value")

    def _build_plots(self):
        """
        Add a plot of the mission
        """

        x_name = self._x_widget.value
        y_name = self._y_widget.value

        for name in self.missions:
            if self._fig is None:
                self._fig = go.Figure()
            # pylint: disable=invalid-name # that's a common naming
            x = self.missions[name][x_name]
            # pylint: disable=invalid-name # that's a common naming
            y = self.missions[name][y_name]

            scatter = go.Scatter(x=x, y=y, mode="lines", name=name)

            self._fig.add_trace(scatter)

            self._fig = go.FigureWidget(self._fig)

        self._fig.update_layout(
            title_text="Mission", title_x=0.5, xaxis_title=x_name, yaxis_title=y_name
        )

    # pylint: disable=unused-argument  # args has to be there for observe() to work
    def display(self, change=None) -> display:
        """
        Display the user interface
        :return the display object
        """
        clear_output(wait=True)
        self._update_plots()
        toolbar = widgets.HBox(
            [widgets.Label(value="x:"), self._x_widget, widgets.Label(value="y:"), self._y_widget]
        )
        # pylint: disable=invalid-name # that's a common naming
        ui = widgets.VBox([toolbar, self._fig])
        return display(ui)

    def _update_plots(self):
        """
        Update the plots
        """
        self._fig = None
        self._build_plots()



from .sub_components.compute_fuel_mass import compute_fuel_mission


def compute_performance(owe, payload, mission_range, tsfc, l_d, cruise_speed):
    """
    Gather all the performances sub-functions in the main function

    :param owe: the structural mass, in kg
    :param payload: the payload mass, in kg
    :param mission_range: the mission range, in m
    :param tsfc: the thrust specific fuel consumption, in kg/N/s
    :param l_d: the lift-to-drag ratio in cruise conditions, no unit
    :param cruise_speed: Cruise speed, in m/s

    :return mission_fuel: the fuel consumed during the designated mission, in kg
    """

    # Let's start by computing the fuel consumed during the mission
    mission_fuel = compute_fuel_mission(
        owe=owe,
        payload=payload,
        mission_range=mission_range,
        tsfc=tsfc,
        l_d=l_d,
        cruise_speed=cruise_speed,
    )

    return mission_fuel



"""This module gathers key FAST-OAD classes and functions for convenient import."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=unused-import
# flake8: noqa


from fastoad.cmd.api import (
    get_plugin_information,
    generate_notebooks,
    evaluate_problem,
    generate_configuration_file,
    generate_inputs,
    list_modules,
    list_variables,
    optimization_viewer,
    optimize_problem,
    variable_viewer,
    write_n2,
    write_xdsm,
)

from fastoad.io.configuration import FASTOADProblemConfigurator
from fastoad.openmdao.problem import FASTOADProblem

from fastoad.io import DataFile
from fastoad.openmdao.variables import Variable, VariableList

from fastoad.model_base import Atmosphere, AtmosphereSI, FlightPoint

from fastoad.module_management.service_registry import (
    RegisterOpenMDAOSystem,
    RegisterPropulsion,
    RegisterSpecializedService,
    RegisterSubmodel,
)
from fastoad.model_base.propulsion import IOMPropulsionWrapper
from fastoad.openmdao.validity_checker import ValidityDomainChecker

from fastoad.model_base.datacls import MANDATORY_FIELD
from fastoad.models.performances.mission.segments.base import (
    AbstractFlightSegment,
    AbstractTimeStepFlightSegment,
    AbstractManualThrustSegment,
    AbstractFixedDurationSegment,
    AbstractRegulatedThrustSegment,
)

from fastoad.gui.mission_viewer import MissionViewer
from fastoad.gui.optimization_viewer import OptimizationViewer
from fastoad.gui.variable_viewer import VariableViewer
from fastoad.gui.analysis_and_plots import (
    aircraft_geometry_plot,
    drag_polar_plot,
    mass_breakdown_bar_plot,
    mass_breakdown_sun_plot,
    wing_geometry_plot,
)



import openmdao.api as om

from ...OpenMDAO.performance.sub_components.compute_fuel_mass import ComputeFuelMass

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.fast_oad.performance")
class ComputePerformance(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_fuel_mass", subsys=ComputeFuelMass(), promotes=["*"])



import numpy as np
import scipy.constants as sc

import openmdao.api as om
from fastoad.model_base import Atmosphere


class ComputeLiftToDragRatio(om.ExplicitComponent):
    """
    Computes the lift to drag ratio considering a lift equilibrium in cruise and a simple
    quadratic model.
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(
            name="cruise_altitude", units="m", val=np.nan
        )  # For a float or int, shape don't have to be
        # provided
        self.add_input(name="cruise_speed", units="m/s", val=np.nan)
        self.add_input(
            name="profile_drag_coefficient", val=np.nan
        )  # When the quantity does not have a unit, the units
        # field doesn't need to be filled
        self.add_input(name="induced_drag_coefficient", val=np.nan)
        self.add_input(name="mtow", units="kg", val=np.nan)
        self.add_input(
            name="wing_area", units="m**2", val=np.nan
        )  # OpenMDAO understands the multiplication/division
        # of units in between them hence why the m**2 is understood

        # Defining the output(s)

        self.add_output(name="l_d_ratio")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        cruise_altitude = inputs["cruise_altitude"]
        cruise_speed = inputs["cruise_speed"]
        cd0 = inputs["profile_drag_coefficient"]
        k = inputs["induced_drag_coefficient"]
        mtow = inputs["mtow"]
        wing_area = inputs["wing_area"]

        # Air density at sea-level, to compute it, we will use the Atmosphere
        # model available in FAST-OAD, so we will create an Atmosphere instance
        # using the cruise altitude and extract its density attribute
        atm = Atmosphere(altitude=cruise_altitude, altitude_in_feet=False)
        rho = atm.density

        # Computation of the cruise lift coefficient using a simple equilibrium
        cl = (mtow * sc.g) / (0.5 * rho * cruise_speed ** 2.0 * wing_area)

        # Computation of the cruise drag coefficient using the simple quadratic model
        cd = cd0 + k * cl ** 2

        # Computation of the ratio
        l_d = cl / cd

        outputs["l_d_ratio"] = l_d



"""
Conversion from OpenMDAO variables to XPath
"""

#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2021 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import IO, Sequence, Set, Union

import numpy as np

from fastoad.io.xml.exceptions import (
    FastXpathTranslatorDuplicates,
    FastXpathTranslatorInconsistentLists,
    FastXpathTranslatorVariableError,
    FastXpathTranslatorXPathError,
)


class VarXpathTranslator:
    """
    Allows to convert OpenMDAO variable names from and to XPath, using a provided conversion table.

    At instantiation, user can provide (as keyword arguments only):
     - variable_names and xpaths (see :meth:`set`)
     - translation file (see :meth:`read_translation_table`)
    """

    def __init__(
        self,
        *,
        variable_names: Sequence[str] = None,
        xpaths: Sequence[str] = None,
        source: Union[str, IO] = None
    ):
        if variable_names is not None and xpaths is not None:
            self.set(variable_names, xpaths)

        if source is not None:
            self.read_translation_table(source)

    def set(self, variable_names: Sequence[str], xpaths: Sequence[str]):
        """
        Sets the "conversion table", i.e. two lists where each element matches the other with same
        index. Provided lists must have the same length.

        :param variable_names: List of OpenMDAO variable names
        :param xpaths: List of XML Paths
        """
        if len(variable_names) != len(xpaths):
            raise FastXpathTranslatorInconsistentLists(
                "lists var_names and xpaths have not the same length (%i and %i)"
                % (len(variable_names), len(xpaths))
            )

        # check duplicate variable names
        dupe_vars = self._get_duplicates(variable_names)
        if dupe_vars:
            raise FastXpathTranslatorDuplicates(
                "Following variable names are provided more than once: %s" % dupe_vars, dupe_vars
            )

        # check duplicate XPaths
        dupe_xpaths = self._get_duplicates(xpaths)
        if dupe_xpaths:
            raise FastXpathTranslatorDuplicates(
                "Following variable names are provided more than once: %s" % dupe_xpaths,
                dupe_xpaths,
            )

        self._variable_names = list(variable_names)
        self._xpaths = list(xpaths)

    def read_translation_table(self, source: Union[str, IO]):
        """
        Reads a file that sets how OpenMDAO variable are matched to XML Path.
        Provided file should have 2 comma-separated columns:
         - first one with OpenMDAO names
         - second one with their matching XPath

        :param source:
        """

        arr = np.genfromtxt(source, dtype=str, delimiter=",", autostrip=True)
        self.set(arr[:, 0], arr[:, 1])

    @property
    def variable_names(self) -> Sequence[str]:
        """List of variable names as set in :meth:`set`"""
        return self._variable_names

    @property
    def xpaths(self) -> Sequence[str]:
        """List of XPaths as set in :meth:`set`"""
        return self._xpaths

    def get_xpath(self, var_name: str) -> str:
        """

        :param var_name: OpenMDAO variable name
        :return: XPath that matches var_name
        :raise FastXpathTranslatorVariableError: if var_name is unknown
        """
        if var_name in self._variable_names:
            i = self._variable_names.index(var_name)
            return self._xpaths[i]
        raise FastXpathTranslatorVariableError(var_name)

    def get_variable_name(self, xpath: str) -> str:
        """

        :param xpath: XML Path
        :return: OpenMDAO variable name that matches xpath
        :raise FastXpathTranslatorXPathError: if xpath is unknown
        """
        if xpath in self._xpaths:
            i = self._xpaths.index(xpath)
            return self._variable_names[i]
        raise FastXpathTranslatorXPathError(xpath)

    @staticmethod
    def _get_duplicates(seq: Sequence) -> Set:
        dupes = set()
        seen = set()
        for elem in seq:
            if elem in seen:
                dupes.add(elem)
            else:
                seen.add(elem)
        return dupes



import openmdao.api as om

from ...OpenMDAO.aerodynamics.sub_components.compute_profile_drag import ComputeProfileDrag
from ...OpenMDAO.aerodynamics.sub_components.compute_induced_drag_coefficient import (
    ComputeInducedDragCoefficient,
)
from ...OpenMDAO.aerodynamics.sub_components.compute_lift_to_drag_ratio import (
    ComputeLiftToDragRatio,
)

import fastoad.api as oad


@oad.RegisterOpenMDAOSystem("tutorial.fast_oad.aerodynamics")
class ComputeAerodynamics(om.Group):
    def setup(self):

        self.add_subsystem(name="compute_profile_drag", subsys=ComputeProfileDrag(), promotes=["*"])
        self.add_subsystem(
            name="compute_induced_drag", subsys=ComputeInducedDragCoefficient(), promotes=["*"]
        )
        self.add_subsystem(
            name="compute_lift_to_drag", subsys=ComputeLiftToDragRatio(), promotes=["*"]
        )



"""Exceptions for module_management package."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Sequence

from fastoad.exceptions import FastError


class FastBundleLoaderDuplicateFactoryError(FastError):
    """
    Raised when trying to register a factory with an already used name.
    """

    def __init__(self, factory_name: str):
        """
        :param factory_name:
        """
        super().__init__('Name "%s" is already used.' % factory_name)
        self.factory_name = factory_name


class FastBundleLoaderUnknownFactoryNameError(FastError):
    """
    Raised when trying to instantiate a component from an unknown factory.
    """

    def __init__(self, factory_name: str):
        """
        :param factory_name:
        """
        super().__init__('"%s" is not registered.' % factory_name)
        self.factory_name = factory_name


class FastBadSystemOptionError(FastError):
    """
    Raised when some option name is not conform to OpenMDAO system definition.
    """

    def __init__(self, identifier, option_names):
        """
        :param identifier: system identifier
        :param option_names: incorrect option names
        """
        super().__init__(
            "OpenMDAO system %s does not accept following option(s): %s"
            % (identifier, option_names)
        )
        self.identifier = identifier
        self.option_names = option_names


class FastIncompatibleServiceClassError(FastError):
    """
    Raised when trying to register as service a class that does not implement
    the specified interface.
    """

    def __init__(self, registered_class: type, service_id: str, base_class: type):
        """
        :param registered_class:
        :param service_id:
        :param base_class: the unmatched interface
        """
        super().__init__(
            'Trying to register %s as service "%s" but it does not inherit from %s'
            % (str(registered_class), service_id, str(base_class))
        )
        self.registered_class = registered_class
        self.service_id = service_id
        self.base_class = base_class


class FastNoSubmodelFoundError(FastError):
    """
    Raised when a submodel is required, but none has been declared.
    """

    def __init__(self, service_id: str):
        """
        :param service_id:
        """
        super().__init__('No submodel found for requirement "%s"' % service_id)
        self.service_id = service_id


class FastTooManySubmodelsError(FastError):
    """
    Raised when several candidates are declared for a required submodel, but
    none has been selected.
    """

    def __init__(self, service_id: str, candidates: Sequence[str]):
        """
        :param service_id:
        :param candidates:
        """
        super().__init__(
            'Submodel requirement "%s" needs a choice among following candidates: %s'
            % (service_id, candidates)
        )
        self.service_id = service_id
        self.candidates = candidates


class FastUnknownSubmodelError(FastError):
    """
    Raised when a submodel identifier is unknown for given required service.
    """

    def __init__(self, service_id: str, submodel_id: str, submodel_ids: List[str]):
        """
        :param service_id:
        :param submodel_id:
        :param submodel_ids:
        """

        msg = '"%s" is not a submodel identifier for requirement "%s"' % (submodel_id, service_id)
        msg += "\nValid identifiers are %s" % submodel_ids
        super().__init__(msg)
        self.service_id = service_id
        self.submodel_id = submodel_id


class FastNoDistPluginError(FastError):
    """Raised when no installed package with FAST-OAD plugin is available."""

    def __init__(self):
        super().__init__("This feature needs plugins, but no plugin available.")


class FastUnknownDistPluginError(FastError):
    """Raised when a distribution name is not found among distribution with FAST-OAD plugins."""

    def __init__(self, dist_name):
        self.dist_name = dist_name
        super().__init__(
            f'No installed package with FAST-OAD plugin found with name "{dist_name}".'
        )


class FastSeveralDistPluginsError(FastError):
    """
    Raised when no distribution name has been specified but several distribution
    with FAST-OAD plugins are available.
    """

    def __init__(self):
        super().__init__(
            "Several installed packages with FAST-OAD plugins are available. "
            "One must be specified."
        )


class FastNoAvailableConfigurationFileError(FastError):
    """Raised when a configuration file is asked, but none is available in plugins."""

    def __init__(self):
        super().__init__("No configuration file provided with currently installed plugins.")


class FastUnknownConfigurationFileError(FastError):
    """Raised when a configuration file is not found for named distribution."""

    def __init__(self, configuration_file, dist_name):
        self.configuration_file = configuration_file
        self.dist_name = dist_name
        super().__init__(
            f'Configuration file "{configuration_file}" not provided with '
            f'installed package "{dist_name}".'
        )


class FastSeveralConfigurationFilesError(FastError):
    """
    Raised when no configuration file has been specified but several configuration files are
    provided with the distribution."""

    def __init__(self, dist_name):
        self.dist_name = dist_name
        super().__init__(
            f'Installed package "{dist_name}" provides several configuration files. '
            "One must be specified."
        )



"""
OpenMDAO component for computation of sizing mission.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import openmdao.api as om

from fastoad.module_management.constants import ModelDomain
from fastoad.module_management.service_registry import RegisterOpenMDAOSystem

_LOGGER = logging.getLogger(__name__)  # Logger for this module


@RegisterOpenMDAOSystem("fastoad.mass_performances.compute_MTOW", domain=ModelDomain.OTHER)
class ComputeMTOW(om.AddSubtractComp):
    """
    Computes MTOW from OWE, design payload and consumed fuel in sizing mission.
    """

    def setup(self):
        self.add_equation(
            "data:weight:aircraft:MTOW",
            [
                "data:weight:aircraft:OWE",
                "data:weight:aircraft:payload",
                "data:weight:aircraft:sizing_onboard_fuel_at_input_weight",
            ],
            units="kg",
        )



"""Classes for climb/descent segments."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from copy import copy
from dataclasses import dataclass
from typing import List, Tuple

import pandas as pd
from scipy.constants import foot, g

from fastoad.model_base import FlightPoint
from .base import AbstractManualThrustSegment
from ..exceptions import FastFlightSegmentIncompleteFlightPoint
from ..util import get_closest_flight_level

_LOGGER = logging.getLogger(__name__)  # Logger for this module


@dataclass
class AltitudeChangeSegment(
    AbstractManualThrustSegment,
    mission_file_keyword="altitude_change",
):
    """
    Computes a flight path segment where altitude is modified with constant speed.

    .. note:: **Setting speed**

        Constant speed may be:

        - constant true airspeed (TAS)
        - constant equivalent airspeed (EAS)
        - constant Mach number

        Target should have :code:`"constant"` as definition for one parameter among
        :code:`true_airspeed`, :code:`equivalent_airspeed` or :code:`mach`.
        All computed flight points will use the corresponding **start** value.
        The two other speed values will be computed accordingly.

        If not "constant" parameter is set, constant TAS is assumed.

    .. note:: **Setting target**

        Target can be an altitude, or a speed:

        - Target altitude can be a float value (in **meters**), or can be set to:

            - :attr:`OPTIMAL_ALTITUDE`: in that case, the target altitude will be the altitude
              where maximum lift/drag ratio is achieved for target speed, depending on current mass.
            - :attr:`OPTIMAL_FLIGHT_LEVEL`: same as above, except that altitude will be rounded to
              the nearest flight level (multiple of 100 feet).

        - For a speed target, as explained above, one value  TAS, EAS or Mach must be
          :code:`"constant"`. One of the two other ones can be set as target.

        In any case, the achieved value will be capped so it respects
        :attr:`maximum_flight_level`.

    """

    time_step: float = 2.0

    #: The maximum allowed flight level (i.e. multiple of 100 feet).
    maximum_flight_level: float = 500.0

    #: Using this value will tell to target the altitude with max lift/drag ratio.
    OPTIMAL_ALTITUDE = "optimal_altitude"  # pylint: disable=invalid-name # used as constant

    #: Using this value will tell to target the nearest flight level to altitude
    #: with max lift/drag ratio.
    OPTIMAL_FLIGHT_LEVEL = "optimal_flight_level"  # pylint: disable=invalid-name # used as constant

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        if target.altitude is not None:
            if isinstance(target.altitude, str):
                # Target altitude will be modified along the process, so we keep track
                # of the original order in target CL, that is not used otherwise.
                target.CL = target.altitude  # pylint: disable=invalid-name
                # let's put a numerical, negative value in target.altitude to
                # ensure there will be no problem in self.get_distance_to_target()
                target.altitude = -1000.0
                self.interrupt_if_getting_further_from_target = False
            else:
                # Target altitude is fixed, back to original settings (in case
                # this instance is used more than once)
                target.CL = None
                self.interrupt_if_getting_further_from_target = True

        atm = self._get_atmosphere_point(start.altitude)
        if target.equivalent_airspeed == self.CONSTANT_VALUE:
            atm.equivalent_airspeed = start.equivalent_airspeed
            start.true_airspeed = atm.true_airspeed
        elif target.mach == self.CONSTANT_VALUE:
            atm.mach = start.mach
            start.true_airspeed = atm.true_airspeed

        return super().compute_from_start_to_target(start, target)

    def get_distance_to_target(
        self, flight_points: List[FlightPoint], target: FlightPoint
    ) -> float:
        current = flight_points[-1]

        # Max flight level is first priority
        max_authorized_altitude = self.maximum_flight_level * 100.0 * foot
        if current.altitude >= max_authorized_altitude:
            return max_authorized_altitude - current.altitude

        if target.CL:
            # Optimal altitude is based on a target Mach number, though target speed
            # may be specified as TAS or EAS. If so, Mach number has to be computed
            # for target altitude and speed.

            # First, as target speed is expected to be set to self.CONSTANT_VALUE for one
            # parameter. Let's get the real value from start point.
            target_speed = copy(target)
            for speed_param in ["true_airspeed", "equivalent_airspeed", "mach"]:
                if isinstance(getattr(target_speed, speed_param), str):
                    setattr(target_speed, speed_param, getattr(flight_points[0], speed_param))

            # Now, let's compute target Mach number
            atm = self._get_atmosphere_point(max(target.altitude, current.altitude))
            if target_speed.equivalent_airspeed:
                atm.equivalent_airspeed = target_speed.equivalent_airspeed
                target_speed.true_airspeed = atm.true_airspeed
            if target_speed.true_airspeed:
                atm.true_airspeed = target_speed.true_airspeed
                target_speed.mach = atm.mach

            # Now we compute optimal altitude
            optimal_altitude = self._get_optimal_altitude(
                current.mass, target_speed.mach, current.altitude
            )
            if target.CL == self.OPTIMAL_ALTITUDE:
                target.altitude = optimal_altitude
            else:  # target.CL == self.OPTIMAL_FLIGHT_LEVEL:
                target.altitude = get_closest_flight_level(optimal_altitude, up_direction=False)

        if target.altitude is not None:
            return target.altitude - current.altitude
        if target.true_airspeed and target.true_airspeed != self.CONSTANT_VALUE:
            return target.true_airspeed - current.true_airspeed
        if target.equivalent_airspeed and target.equivalent_airspeed != self.CONSTANT_VALUE:
            return target.equivalent_airspeed - current.equivalent_airspeed
        if target.mach is not None and target.mach != self.CONSTANT_VALUE:
            return target.mach - current.mach

        raise FastFlightSegmentIncompleteFlightPoint(
            "No valid target definition for altitude change."
        )

    def get_gamma_and_acceleration(self, flight_point: FlightPoint) -> Tuple[float, float]:
        gamma = (flight_point.thrust - flight_point.drag) / flight_point.mass / g
        return gamma, 0.0



"""
OpenMDAO component for time-step computation of missions.
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from enum import Enum
from os import makedirs, path as pth

import numpy as np
from openmdao import api as om

from fastoad.model_base import FlightPoint
from fastoad.model_base.propulsion import IOMPropulsionWrapper
from fastoad.models.performances.mission.openmdao.mission_wrapper import MissionWrapper
from fastoad.models.performances.mission.polar import Polar
from fastoad.models.performances.mission.segments.cruise import BreguetCruiseSegment
from fastoad.module_management.service_registry import RegisterPropulsion

_LOGGER = logging.getLogger(__name__)  # Logger for this module


class MissionComponent(om.ExplicitComponent):
    """
    Computes a mission as specified in mission input file

    Options:
      - propulsion_id: (mandatory) the identifier of the propulsion wrapper.
      - out_file: if provided, a csv file will be written at provided path with all computed
                  flight points.
      - mission_wrapper: the MissionWrapper instance that defines the mission.
      - use_initializer_iteration: During first solver loop, a complete mission computation can
                                   fail or consume useless CPU-time. When activated, this option
                                   ensures the first iteration is done using a simple, dummy,
                                   formula instead of the specified mission.
                                   Set this option to False if you do expect this model to be
                                   computed only once.
      - is_sizing: if True, TOW will be considered equal to MTOW and mission payload will be
                   considered equal to design payload.
      - reference_area_variable: Defines the name of the variable for providing aircraft
                                 reference surface area.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.flight_points = None
        self._engine_wrapper = None
        self._mission_wrapper: MissionWrapper = None
        self._name_provider = None
        self._input_weight_variable_name = ""

    def initialize(self):
        self.options.declare("propulsion_id", default="", types=str)
        self.options.declare("out_file", default="", types=str)
        self.options.declare("use_initializer_iteration", default=True, types=bool)
        self.options.declare("mission_wrapper", types=MissionWrapper)
        self.options.declare("mission_name", types=str)
        self.options.declare("is_sizing", default=False, types=bool)
        self.options.declare(
            "reference_area_variable", default="data:geometry:wing:area", types=str
        )

    def setup(self):
        self._engine_wrapper = self._get_engine_wrapper()
        self._engine_wrapper.setup(self)
        self._mission_wrapper = self.options["mission_wrapper"]
        self._mission_wrapper.setup(self, self.options["mission_name"])

        mission_name = self.options["mission_name"]

        self._name_provider = _get_variable_name_provider(mission_name)
        self._input_weight_variable_name = self._mission_wrapper.get_input_weight_variable_name(
            self.options["mission_name"]
        )

        try:
            self.add_input(self.options["reference_area_variable"], np.nan, units="m**2")
        except ValueError:
            pass

        # Global mission outputs
        self.add_output(
            self._name_provider.NEEDED_BLOCK_FUEL.value,
            units="kg",
            desc=f'Needed fuel to complete mission "{mission_name}", including reserve fuel',
        )
        self.add_output(
            self._name_provider.CONSUMED_FUEL_BEFORE_INPUT_WEIGHT.value,
            units="kg",
            desc=f'consumed fuel quantity before target mass defined for "{mission_name}",'
            f" if any (e.g. TakeOff Weight)",
        )
        if self.options["is_sizing"]:
            self.add_output("data:weight:aircraft:sizing_block_fuel", units="kg")
            self.add_output("data:weight:aircraft:sizing_onboard_fuel_at_input_weight", units="kg")

    def setup_partials(self):
        self.declare_partials(["*"], ["*"], method="fd")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        iter_count = self.iter_count_without_approx
        message_prefix = f"Mission computation - iteration {iter_count:d} : "
        if iter_count == 0 and self.options["use_initializer_iteration"]:
            _LOGGER.info(
                "%sUsing initializer computation. OTHER ITERATIONS NEEDED.", message_prefix
            )
            self._compute_breguet(inputs, outputs)
        else:
            _LOGGER.info("%sUsing mission definition.", message_prefix)
            self._compute_mission(inputs, outputs)

    def _compute_breguet(self, inputs, outputs):
        """
        Computes mission using simple Breguet formula at altitude==100m and Mach 0.1
        (but max L/D ratio is assumed anyway)

        Useful only for initiating the computation.

        :param inputs: OpenMDAO input vector
        :param outputs: OpenMDAO output vector
        """
        propulsion_model = self._engine_wrapper.get_model(inputs)

        high_speed_polar = self._get_initial_polar(inputs)
        distance = np.sum(
            self._mission_wrapper.get_route_ranges(inputs, self.options["mission_name"])
        ).item()

        altitude = 100.0
        cruise_mach = 0.1

        breguet = BreguetCruiseSegment(
            target=FlightPoint(ground_distance=distance),
            propulsion=propulsion_model,
            polar=high_speed_polar,
            use_max_lift_drag_ratio=True,
        )
        start_point = FlightPoint(
            mass=inputs[self._input_weight_variable_name],
            altitude=altitude,
            mach=cruise_mach,
        )
        flight_points = breguet.compute_from(start_point)
        end_point = FlightPoint.create(flight_points.iloc[-1])
        outputs[self._name_provider.NEEDED_BLOCK_FUEL.value] = start_point.mass - end_point.mass

    @staticmethod
    def _get_initial_polar(inputs) -> Polar:
        """
        At computation start, polar may be irrelevant and give a very low lift/drag ratio.

        In that case, this method returns a fake polar that has 10.0 as max lift drag ratio.
        Otherwise, the actual cruise polar is returned.
        """
        high_speed_polar = Polar(
            inputs["data:aerodynamics:aircraft:cruise:CL"],
            inputs["data:aerodynamics:aircraft:cruise:CD"],
        )
        use_minimum_l_d_ratio = False
        try:
            if (
                high_speed_polar.optimal_cl / high_speed_polar.cd(high_speed_polar.optimal_cl)
                < 10.0
            ):
                use_minimum_l_d_ratio = True
        except ZeroDivisionError:
            use_minimum_l_d_ratio = True
        if use_minimum_l_d_ratio:
            # We replace by a polar that has at least 10.0 as max L/D ratio
            high_speed_polar = Polar(np.array([0.0, 0.5, 1.0]), np.array([0.1, 0.05, 1.0]))

        return high_speed_polar

    def _compute_mission(self, inputs, outputs):
        """
        Computes mission using time-step integration.

        :param inputs: OpenMDAO input vector
        :param outputs: OpenMDAO output vector
        """
        propulsion_model = self._engine_wrapper.get_model(inputs)
        reference_area = inputs[self.options["reference_area_variable"]]

        self._mission_wrapper.propulsion = propulsion_model
        self._mission_wrapper.reference_area = reference_area

        # This is the default start point, that can be overridden by using the "start"
        # segment in the mission definition.
        start_flight_point = FlightPoint(
            altitude=0.0, mass=inputs[self._input_weight_variable_name], true_airspeed=0.0
        )

        self.flight_points = self._mission_wrapper.compute(start_flight_point, inputs, outputs)

        # Final ================================================================
        start_of_mission = FlightPoint.create(self.flight_points.iloc[0])
        end_of_mission = FlightPoint.create(self.flight_points.iloc[-1])
        reserve = self._mission_wrapper.get_reserve(
            self.flight_points, self.options["mission_name"]
        )
        reserve_name = self._mission_wrapper.get_reserve_variable_name()
        if reserve_name in outputs:
            outputs[reserve_name] = reserve
        outputs[self._name_provider.NEEDED_BLOCK_FUEL.value] = (
            start_of_mission.mass - end_of_mission.mass + reserve
        )

        outputs[
            self._name_provider.CONSUMED_FUEL_BEFORE_INPUT_WEIGHT.value
        ] = self._mission_wrapper.consumed_fuel_before_input_weight

        if self.options["is_sizing"]:
            outputs["data:weight:aircraft:sizing_block_fuel"] = outputs[
                self._name_provider.NEEDED_BLOCK_FUEL.value
            ]
            outputs["data:weight:aircraft:sizing_onboard_fuel_at_input_weight"] = (
                outputs[self._name_provider.NEEDED_BLOCK_FUEL.value]
                - self._mission_wrapper.consumed_fuel_before_input_weight
            )

        def as_scalar(value):
            if isinstance(value, np.ndarray):
                return value.item()
            return value

        self.flight_points = self.flight_points.applymap(as_scalar)
        rename_dict = {
            field_name: f"{field_name} [{unit}]"
            for field_name, unit in FlightPoint.get_units().items()
        }
        self.flight_points.rename(columns=rename_dict, inplace=True)

        if self.options["out_file"]:
            makedirs(pth.dirname(self.options["out_file"]), exist_ok=True)
            self.flight_points.to_csv(self.options["out_file"])

    def _get_engine_wrapper(self) -> IOMPropulsionWrapper:
        """
        Overloading this method allows to define the engine without relying on the propulsion
        option.

        (useful for tests)

        :return: the engine wrapper instance
        """
        return RegisterPropulsion.get_provider(self.options["propulsion_id"])


def _get_variable_name_provider(mission_name):
    """Factory for enum class that provide mission variable names."""

    def get_variable_name(suffix):
        return f"data:mission:{mission_name}:{suffix}"

    class VariableNames(Enum):
        """Enum with mission-related variable names."""

        ZFW = get_variable_name("ZFW")
        PAYLOAD = get_variable_name("payload")
        BLOCK_FUEL = get_variable_name("block_fuel")
        NEEDED_BLOCK_FUEL = get_variable_name("needed_block_fuel")
        CONSUMED_FUEL_BEFORE_INPUT_WEIGHT = get_variable_name("consumed_fuel_before_input_weight")

    return VariableNames



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fastoad',
 'fastoad._utils',
 'fastoad._utils.resource_management',
 'fastoad.cmd',
 'fastoad.cmd.resources',
 'fastoad.configurations',
 'fastoad.gui',
 'fastoad.io',
 'fastoad.io.configuration',
 'fastoad.io.configuration.resources',
 'fastoad.io.xml',
 'fastoad.io.xml.resources',
 'fastoad.model_base',
 'fastoad.models',
 'fastoad.models.performances',
 'fastoad.models.performances.mission',
 'fastoad.models.performances.mission.mission_definition',
 'fastoad.models.performances.mission.mission_definition.mission_builder',
 'fastoad.models.performances.mission.mission_definition.resources',
 'fastoad.models.performances.mission.openmdao',
 'fastoad.models.performances.mission.openmdao.resources',
 'fastoad.models.performances.mission.segments',
 'fastoad.module_management',
 'fastoad.notebooks',
 'fastoad.notebooks.01_Quick_start',
 'fastoad.notebooks.01_Quick_start.data',
 'fastoad.notebooks.01_Quick_start.img',
 'fastoad.notebooks.01_Quick_start.modules',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.data',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.img',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.aerodynamics',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.geometry',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.mass',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.performance',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.performance.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.FAST-OAD.update_mtow',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.aerodynamics',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.aerodynamics.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.geometry',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.geometry.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.mass',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.mass.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.performance',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.performance.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.OpenMDAO.update_mtow',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.aerodynamics',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.aerodynamics.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.geometry',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.geometry.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.mass',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.mass.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.performance',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.performance.sub_components',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.pure_python.update_mtow',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.modules.python_for_scipy',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.working_folder',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.working_folder.MDA_discipline',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.working_folder.MDA_global_group',
 'fastoad.notebooks.02_From_Python_to_FAST-OAD.working_folder.MDO_MTOW',
 'fastoad.openmdao',
 'fastoad.openmdao.resources',
 'fastoad.openmdao.variables']

package_data = \
{'': ['*']}

install_requires = \
['Deprecated>=1.2.13,<2.0.0',
 'aenum>=3.1.0,<4.0.0',
 'click>=8.0.3,<9.0.0',
 'ensure>=1.0.0,<2.0.0',
 'ipopo>=1.0.0,<2.0.0',
 'ipysheet>=0.5.0,<1',
 'ipywidgets>=7.7.0,<8.0.0',
 'jsonschema>=3.2.0,<5',
 'jupyter-client!=7.0.0,!=7.0.1,!=7.0.2,!=7.0.3,!=7.0.4,!=7.0.5',
 'jupyterlab>=3.3.0,<4.0.0',
 'lxml>=4.6.5,<5.0.0',
 'notebook>=6.0,<7.0',
 'numpy>=1.21.0,<2.0.0',
 'openmdao>=3.10,<=3.21',
 'pandas>=1.1.0,<2.0.0',
 'plotly>=5.0.0,<6.0.0',
 'ruamel.yaml>=0.15.78,<0.18',
 'scipy>=1.4.1,<2.0.0',
 'stdatm<1.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tomlkit>=0.5.3,<1',
 'wop>=2.2.0,<3.0.0']

extras_require = \
{':python_version < "3.10"': ['importlib-metadata>=4.2,<5.0'],
 'mpi4py': ['mpi4py>=3,<4']}

entry_points = \
{'console_scripts': ['fast-oad = fastoad.cmd.cli:fast_oad',
                     'fastoad = fastoad.cmd.cli:fast_oad'],
 'fastoad.plugins': ['bundled = fastoad']}

setup_kwargs = {
    'name': 'fast-oad-core',
    'version': '1.4.0',
    'description': 'FAST-OAD is a framework for performing rapid Overall Aircraft Design',
    'long_description': '![Tests](https://github.com/fast-aircraft-design/FAST-OAD/workflows/Tests/badge.svg)\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/9691f1d1430c45cf9c94bc342b4c6122)](https://www.codacy.com/gh/fast-aircraft-design/FAST-OAD?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=fast-aircraft-design/FAST-OAD&amp;utm_campaign=Badge_Grade)\n[![codecov](https://codecov.io/gh/fast-aircraft-design/FAST-OAD/branch/master/graph/badge.svg)](https://codecov.io/gh/fast-aircraft-design/FAST-OAD)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![License: GPL v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)\n\n[![Documentation Status](https://readthedocs.org/projects/fast-oad/badge/?version=stable)](https://fast-oad.readthedocs.io/)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/fast-aircraft-design/FAST-OAD.git/latest-release?urlpath=lab%2Ftree%2Fsrc%2Ffastoad%2Fnotebooks)\n\n\nFAST-OAD: Future Aircraft Sizing Tool - Overall Aircraft Design\n===============================================================\n\nFAST-OAD is a framework for performing rapid Overall Aircraft Design.\n\nIt proposes multi-disciplinary analysis and optimisation by relying on\nthe [OpenMDAO framework](https://openmdao.org/).\n\nFAST-OAD allows easy switching between models for a same discipline, and\nalso adding/removing/developing models to match the need of your study.\n\nMore details can be found in the [official documentation](https://fast-oad.readthedocs.io/).\n\n> **Important notice:**\n>\n> Since version 1.3.0, FAST-OAD models for commercial transport aircraft have moved in \n> package  \n> [FAST-OAD-CS25](https://pypi.org/project/fast-oad-cs25/). This package is installed along with \n> FAST-OAD, to keep backward compatibility.\n> \n> Keep in mind that any update of these models will now come through new releases of FAST-OAD-CS25.\n> \n> To get FAST-OAD without these models, you may install\n> [FAST-OAD-core](https://pypi.org/project/fast-oad-core/).\n> \n> :warning: Upgrading from an earlier version than 1.3 may break the `fastoad` command line (no \n> impact on PythonAPI). See [this issue](https://github.com/fast-aircraft-design/FAST-OAD/issues/425)\n> for details and fix.\n\nWant to try quickly?\n--------------------\nYou can run FAST-OAD tutorials **without installation** using our\n[Binder-hosted Jupyter notebooks](https://mybinder.org/v2/gh/fast-aircraft-design/FAST-OAD.git/latest-release?filepath=src%2Ffastoad%2Fnotebooks).\n\n\nInstall\n-------\n\n**Prerequisite**:FAST-OAD needs at least **Python 3.7.0**.\n\nIt is recommended (but not required) to install FAST-OAD in a virtual\nenvironment ([conda](https://docs.conda.io/en/latest/),\n[venv](https://docs.python.org/3.7/library/venv.html), ...)\n\nOnce Python is installed, FAST-OAD can be installed using pip.\n\n> **Note**: If your network uses a proxy, you may have to do [some\n> settings](https://pip.pypa.io/en/stable/user_guide/#using-a-proxy-server)\n> for pip to work correctly\n\nYou can install the latest version with this command:\n\n``` {.bash}\n$ pip install --upgrade fast-oad\n```\n\nor, if you want the minimum installation without the CS25-related models:\n\n``` {.bash}\n$ pip install --upgrade fast-oad-core\n```\n',
    'author': 'Christophe DAVID',
    'author_email': 'christophe.david@onera.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/fast-aircraft-design/FAST-OAD',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import numpy as np

import openmdao.api as om


class ComputeWingArea(om.ExplicitComponent):
    """
    Computes the wing area based on the provided MTOW and wing loading
    """

    def setup(self):

        # Defining the input(s)

        self.add_input(name="mtow", units="kg", val=np.nan)
        self.add_input(name="wing_loading", units="kg/m**2", val=np.nan)

        # Defining the output(s)

        self.add_output(name="wing_area", units="m**2")

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):

        # Assigning the input to local variable for clarity
        mtow = inputs["mtow"]

        wing_loading = inputs["wing_loading"]

        # Computation of the wing area
        wing_area = mtow / wing_loading

        outputs["wing_area"] = wing_area



"""
Module for custom Exception classes
"""


#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.


class FastError(Exception):
    """
    Base Class for exceptions related to the FAST framework.
    """


class NoSetupError(FastError):
    """
    No Setup Error.

    This exception indicates that a setup of the OpenMDAO instance has not been done, but was
    expected to be.
    """


class XMLReadError(FastError):
    """
    XML file read Error.

    This exception indicates that an error occurred when reading an xml file.
    """


class FastUnknownEngineSettingError(FastError):
    """
    Raised when an unknown engine setting code has been encountered
    """


class FastUnexpectedKeywordArgument(FastError):
    """
    Raised when an instantiation is done with an incorrect keyword argument.
    """

    def __init__(self, bad_keyword):
        super().__init__("Unexpected keyword argument: %s" % bad_keyword)
        self.bad_keyword = bad_keyword



"""Command Line Interface."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os.path as pth

import click
import tabulate

import fastoad
from fastoad.cmd.cli_utils import (
    manage_overwrite,
    out_file_option,
    overwrite_option,
)
from fastoad.cmd.exceptions import FastNoAvailableNotebookError
from fastoad.module_management.exceptions import (
    FastNoAvailableConfigurationFileError,
    FastNoDistPluginError,
    FastSeveralConfigurationFilesError,
    FastSeveralDistPluginsError,
    FastUnknownConfigurationFileError,
    FastUnknownDistPluginError,
)
from . import api

NOTEBOOK_FOLDER_NAME = "FAST-OAD_notebooks"


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(fastoad.__version__, "-v", "--version")
def fast_oad():
    """FAST-OAD main program"""


@fast_oad.command(name="plugin_info")
def plugin_info():
    """Provides list of installed FAST-OAD plugins."""
    api.get_plugin_information(print_data=True)


@fast_oad.command(name="gen_conf")
@click.argument("conf_file", nargs=1)
@click.option(
    "-p",
    "--from_package",
    nargs=1,
    help="Name of installed package that provides the sample configuration file.",
)
@click.option("-s", "--source", nargs=1, help="Name of source configuration file.")
@overwrite_option
def gen_conf(conf_file, from_package, source, force):
    """
    Generate a sample configuration file with given argument as name.

    Option "--from_package" has to be used if several FAST-OAD plugins are available.

    Option "--source_name" has to be used if the targeted plugin provides several sample
    configuration files.

    Use "fastoad plugin_info" to get information about available plugins and sample
    configuration files.
    """
    try:
        manage_overwrite(
            api.generate_configuration_file,
            configuration_file_path=conf_file,
            overwrite=force,
            distribution_name=from_package,
            sample_file_name=source,
        )
    except (
        FastNoDistPluginError,
        FastSeveralDistPluginsError,
        FastUnknownDistPluginError,
        FastSeveralConfigurationFilesError,
        FastUnknownConfigurationFileError,
        FastNoAvailableConfigurationFileError,
    ) as exc:
        click.echo(exc.args[0])


@fast_oad.command(name="gen_inputs")
@click.argument("conf_file", nargs=1)
@click.argument("source_file", nargs=1, required=False)
@overwrite_option
@click.option(
    "--legacy", is_flag=True, help="To be used if the source XML file is in legacy format."
)
def gen_inputs(conf_file, source_file, force, legacy):
    """
    Generate the input file (specified in the configuration file) with needed variables.

    \b
    Examples:
    ---------
    # For the problem defined in conf_file.yml, generates the input file with default
    # values (when default values are defined):
        fastoad gen_inputs conf_file.yml

    \b
    # Same as above, except that values are taken from some_file.xml when possible:
        fastoad gen_inputs conf_file.yml some_file.xml

    \b
    # Same as above, some_file.xml is formatted with the legacy FAST schema
        fastoad gen_inputs conf_file.yml some_file.xml --legacy
    """
    schema = "legacy" if legacy else "native"
    manage_overwrite(
        api.generate_inputs,
        configuration_file_path=conf_file,
        source_path=source_file,
        source_path_schema=schema,
        overwrite=force,
    )


@fast_oad.command(name="list_modules")
@out_file_option
@click.option("-v", "--verbose", is_flag=True, help="Shows detailed information for each system.")
@click.argument("source_path", nargs=-1)
def list_modules(out_file, force, verbose, source_path):
    """
    Provide the identifiers of available systems.

    SOURCE_PATH argument can be a configuration file, or a list of folders where
    custom modules are declared.
    """
    # If a configuration file or a single path is provided make sure it is sent as a
    # string not a list
    if len(source_path) == 1:
        source_path = source_path[0]

    if manage_overwrite(
        api.list_modules,
        source_path=source_path,
        out=out_file,
        overwrite=force,
        verbose=verbose,
    ):
        print("\nDone. Use --verbose (-v) option for detailed information.")


@fast_oad.command(name="list_variables")
@click.argument("conf_file", nargs=1)
@out_file_option
@click.option(
    "--format",
    "table_format",
    default="grid",
    show_default=True,
    help=f"format of the list. Available options are {['var_desc'] + tabulate.tabulate_formats}. "
    '"var_desc" is the variable_descriptions.txt format. Other formats are part of the '
    "tabulate package.",
)
def list_variables(conf_file, out_file, force, table_format):
    """List the variables of the problem defined in CONF_FILE."""
    manage_overwrite(
        api.list_variables,
        configuration_file_path=conf_file,
        out=out_file,
        overwrite=force,
        tablefmt=table_format,
    )


@fast_oad.command(name="n2")
@click.argument("conf_file", nargs=1)
@click.argument("n2_file", nargs=1, default="n2.html", required=False)
@overwrite_option
def write_n2(conf_file, n2_file, force):
    """
    Write an HTML file that shows the N2 diagram of the problem defined in CONF_FILE.

    The name of generated file is `n2.html`, or the given name for argument N2_FILE.
    """
    manage_overwrite(
        api.write_n2,
        configuration_file_path=conf_file,
        n2_file_path=n2_file,
        overwrite=force,
    )


@fast_oad.command(name="xdsm")
@click.argument("conf_file", nargs=1)
@click.argument("xdsm_file", nargs=1, default="xdsm.html", required=False)
@overwrite_option
@click.option("--depth", default=2, show_default=True, help="Depth of analysis.")
@click.option(
    "--server",
    help="URL of WhatsOpt server. For advanced users only.",
)
def write_xdsm(conf_file, xdsm_file, depth, server, force):
    """
    Write an HTML file that shows the XDSM diagram of the problem defined in CONF_FILE.

    The name of generated file is `xdsm.html`, or the given name for argument XDSM_FILE.
    """
    manage_overwrite(
        api.write_xdsm,
        configuration_file_path=conf_file,
        xdsm_file_path=xdsm_file,
        overwrite=force,
        depth=depth,
        wop_server_url=server,
    )


@fast_oad.command(name="eval")
@click.argument("conf_file", nargs=1)
@overwrite_option
def evaluate(conf_file, force):
    """Run the analysis for problem defined in CONF_FILE."""
    manage_overwrite(
        api.evaluate_problem,
        filename_func=lambda pb: pb.output_file_path,
        configuration_file_path=conf_file,
        overwrite=force,
    )


@fast_oad.command(name="optim")
@click.argument("conf_file", nargs=1)
@overwrite_option
def optimize(conf_file, force):
    """Run the optimization for problem defined in CONF_FILE."""
    manage_overwrite(
        api.optimize_problem,
        filename_func=lambda pb: pb.output_file_path,
        configuration_file_path=conf_file,
        overwrite=force,
    )


@fast_oad.command(name="notebooks")
@click.argument("path", nargs=1, default=".", required=False)
@click.option(
    "-p", "--from_package", nargs=1, help="Will use only notebooks from this installed package."
)
def create_notebooks(path, from_package):
    """
    Creates a FAST-OAD_notebooks/ folder with pre-configured Jupyter notebooks.

    If PATH is given, FAST-OAD_notebooks/ will be created in that folder.



    IMPORTANT: Please note that all content of an existing FAST-OAD_notebooks/ will be overwritten.
    """

    root_target_path = pth.abspath(pth.join(path, NOTEBOOK_FOLDER_NAME))
    try:
        if manage_overwrite(
            api.generate_notebooks,
            destination_path=root_target_path,
            overwrite=False,
            distribution_name=from_package,
        ):
            # Give info for running Jupyter
            click.echo("You may now run Jupyter with:")
            click.echo(f'   jupyter lab "{root_target_path}"')
    except (
        FastNoDistPluginError,
        FastUnknownDistPluginError,
        FastNoAvailableNotebookError,
    ) as exc:
        click.echo(exc.args[0])


if __name__ == "__main__":
    fast_oad()



"""Base classes for simulating flight segments."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Type

import numpy as np
import pandas as pd
from aenum import Enum, extend_enum
from deprecated import deprecated
from scipy.constants import g
from scipy.optimize import root_scalar
from stdatm import AtmosphereSI

from fastoad.constants import EngineSetting
from fastoad.model_base import FlightPoint
from fastoad.model_base.datacls import MANDATORY_FIELD
from fastoad.model_base.propulsion import IPropulsion
from fastoad.models.performances.mission.polar import Polar
from ..base import IFlightPart
from ..exceptions import FastFlightSegmentIncompleteFlightPoint

_LOGGER = logging.getLogger(__name__)  # Logger for this module

DEFAULT_TIME_STEP = 0.2


class SegmentDefinitions(Enum):
    """
    Class that associates segment names (mission file keywords) and their implementation.
    """

    @classmethod
    def add_segment(cls, segment_name: str, segment_class: Type["AbstractFlightSegment"]):
        """
        Adds a segment definition.

        :param segment_name: segment names (mission file keyword)
        :param segment_class: segment implementation (derived of :class:`~FlightSegment`)
        """
        if issubclass(segment_class, AbstractFlightSegment):
            extend_enum(cls, segment_name, segment_class)
        else:
            raise RuntimeWarning(
                f'"{segment_name}" is ignored as segment name because its associated class '
                "does not derive from FlightSegment."
            )

    @classmethod
    def get_segment_class(cls, segment_name) -> Optional[Type["AbstractFlightSegment"]]:
        """
        Provides the segment implementation for provided name.

        :param segment_name:
        :return: the segment implementation (derived of :class:`~FlightSegment`),
                 or None if segment_name is not known
        """
        try:
            enum = cls[segment_name]
        except KeyError:
            enum = None

        return enum.value if enum else None


@dataclass
class AbstractFlightSegment(IFlightPart, ABC):
    """
    Base class for flight path segment.

    As a dataclass, attributes can be set at instantiation.

    When subclassing this class, the attribute "mission_file_keyword" can be set,
    so that the segment can be used in mission file definition with this keyword:

        >>> class NewSegment(AbstractFlightSegment, mission_file_keyword="new_segment")
        >>>     ...

    Then in mission definition:

    .. code-block:: yaml

        phases:
            my_phase:
                parts:
                    - segment: new_segment

    .. Important::

        :meth:`compute_from` is the method to call to achieve the segment computation.

        However, when subclassing, the method to overload is :meth:`compute_from_start_to_target`.
        Generic reprocessing of start and target flight points is done in :meth:`compute_from`
        before calling :meth:`compute_from_start_to_target`

    """

    #: A FlightPoint instance that provides parameter values that should all be reached at the
    #: end of :meth:`~fastoad.models.performances.mission.segments.base.FlightSegment.compute_from`.
    #: Possible parameters depend on the current segment. A parameter can also be set to
    #: :attr:`~fastoad.models.performances.mission.segments.base.FlightSegment.CONSTANT_VALUE`
    #: to tell that initial value should be kept during all segment.
    target: FlightPoint = MANDATORY_FIELD

    # the `target` field above will be overloaded by a property, using the hidden value below:
    _target: FlightPoint = field(default=MANDATORY_FIELD, init=False)

    #: The temperature offset for ISA atmosphere model.
    isa_offset: float = 0.0

    #: Using this value will tell to keep the associated parameter constant.
    CONSTANT_VALUE = "constant"  # pylint: disable=invalid-name # used as constant

    # To be noted: this one is not a dataclass field, but an actual class attribute
    _attribute_units = dict(reference_area="m**2", time_step="s")

    @abstractmethod
    def compute_from_start_to_target(self, start, target) -> pd.DataFrame:
        """
        Here should come the implementation for computing flight points
        between start and target flight points.

        :param start:
        :param target: Definition of segment target
        :return: a pandas DataFrame where column names match fields of
                 :class:`~fastoad.model_base.flight_point.FlightPoint`
        """

    @classmethod
    def __init_subclass__(cls, *, mission_file_keyword="", attribute_units: Dict[str, str] = None):
        if mission_file_keyword:
            SegmentDefinitions.add_segment(mission_file_keyword, cls)
        cls._attribute_units = cls._attribute_units.copy()
        if attribute_units:
            cls._attribute_units.update(attribute_units)

        # We want to have self.target as a property to ensure it gets always "scalarized".
        # But properties and dataclasses do not mix very well. It would be possible to
        # declare "target" as property, though it is a dataclass field, but it would then
        # be considered as a field with defined default value (the default being the property
        # object XD ). And since it is followed by fields without default, Python complains.
        #
        # The solution is to define the property afterwards, and since we are in an abstract class,
        # it can be done when subclassing.
        def _get_target(self) -> FlightPoint:
            return self._target

        def _set_target(self, value: FlightPoint):
            value.scalarize()
            # Initializing self._target elsewhere give bad results. Since target is mandatory
            # in constructor, this initialization will happen.
            self._target = value

        cls.target = property(_get_target, _set_target)

    @classmethod
    def get_attribute_unit(cls, attribute_name: str) -> str:
        """Returns unit for specified attribute."""
        return cls._attribute_units.get(attribute_name)

    def compute_from(self, start: FlightPoint) -> pd.DataFrame:
        """
        Computes the flight path segment from provided start point.

        Computation ends when target is attained, or if the computation stops getting
        closer to target.
        For instance, a climb computation with too low thrust will only return one
        flight point, that is the provided start point.

        .. Important::

            When subclasssing, if you need to overload :meth:`compute_from`, you should consider
            overriding :meth:`_compute_from` instead. Therefore, you will take benefit of the
            preprocessing of start and target flight points that is done in :meth:`compute_from`


        :param start: the initial flight point, defined for `altitude`, `mass` and speed
                      (`true_airspeed`, `equivalent_airspeed` or `mach`). Can also be
                      defined for `time` and/or `ground_distance`.
        :return: a pandas DataFrame where column names match fields of
                 :class:`~fastoad.model_base.flight_point.FlightPoint`
        """
        # Let's ensure we do not modify the original definitions of start and target
        # during the process
        start_copy = deepcopy(start)
        if start_copy.altitude is not None:
            try:
                self.complete_flight_point(start_copy)
            except FastFlightSegmentIncompleteFlightPoint:
                pass
        start_copy.scalarize()

        target_copy = self._target.make_absolute(start_copy)
        target_copy.scalarize()

        if start_copy.time is None:
            start_copy.time = 0.0
        if start_copy.ground_distance is None:
            start_copy.ground_distance = 0.0

        flight_points = self.compute_from_start_to_target(start_copy, target_copy)

        return flight_points

    def complete_flight_point(self, flight_point: FlightPoint):
        """
        Computes data for provided flight point.

        Assumes that it is already defined for time, altitude, mass,
        ground distance and speed (TAS, EAS, or Mach).

        :param flight_point: the flight point that will be completed in-place
        """

        self._complete_speed_values(flight_point)

    @staticmethod
    def complete_flight_point_from(flight_point: FlightPoint, source: FlightPoint):
        """
        Sets undefined values in `flight_point` using the ones from `source`.

        The particular case of speeds is taken into account: if at least one speed parameter is
        defined, all other speed parameters are considered defined, because they will be
        deduced when needed.

        :param flight_point:
        :param source:
        """
        speed_fields = {
            "true_airspeed",
            "equivalent_airspeed",
            "calibrated_airspeed",
            "mach",
            "unitary_reynolds",
        }.intersection(flight_point.get_field_names())
        other_fields = list(set(flight_point.get_field_names()) - speed_fields)

        speeds_are_missing = np.all([getattr(flight_point, name) is None for name in speed_fields])

        for field_name in other_fields + speeds_are_missing * list(speed_fields):
            if getattr(flight_point, field_name) is None and not source.is_relative(field_name):
                setattr(flight_point, field_name, getattr(source, field_name))

    def _complete_speed_values(
        self, flight_point: FlightPoint, raise_error_on_missing_speeds=True
    ) -> bool:
        """
        Computes consistent values between TAS, EAS and Mach, assuming one of them is defined.
        """
        atm = self._get_atmosphere_point(flight_point.altitude)

        if flight_point.true_airspeed is None:
            if flight_point.mach is not None:
                atm.mach = flight_point.mach
            elif flight_point.equivalent_airspeed is not None:
                atm.equivalent_airspeed = flight_point.equivalent_airspeed
            elif raise_error_on_missing_speeds:
                raise FastFlightSegmentIncompleteFlightPoint(
                    "Flight point should be defined for true_airspeed, "
                    "equivalent_airspeed, or mach."
                )
            else:
                return False
            flight_point.true_airspeed = atm.true_airspeed
        else:
            atm.true_airspeed = flight_point.true_airspeed

        flight_point.mach = atm.mach
        flight_point.equivalent_airspeed = atm.equivalent_airspeed
        return True

    def _get_atmosphere_point(self, altitude: float) -> AtmosphereSI:
        """
        Convenience method to ensure used atmosphere model is initiated with :attr:`delta_isa`.

        :param altitude: in meters
        :return: AtmosphereSI instantiated from provided altitude and :attr:`delta_isa`
        """
        return AtmosphereSI(altitude, self.isa_offset)


@dataclass
class AbstractTimeStepFlightSegment(
    AbstractFlightSegment,
    ABC,
):
    """
    Base class for time step computation flight segments.

    This class implements the time computation. For this computation to work, subclasses must
    implement abstract methods :meth:`get_distance_to_target`,
    :meth:`get_gamma_and_acceleration` and :meth:`compute_propulsion`.
    """

    #: A IPropulsion instance that will be called at each time step.
    propulsion: IPropulsion = MANDATORY_FIELD

    #: The Polar instance that will provide drag data.
    polar: Polar = MANDATORY_FIELD

    #: The reference area, in m**2.
    reference_area: float = MANDATORY_FIELD

    #: Used time step for computation (actual time step can be lower at some particular times of
    #: the flight path).
    time_step: float = DEFAULT_TIME_STEP

    #: Minimum and maximum authorized altitude values. If computed altitude gets beyond these
    #: limits, computation will be interrupted and a warning message will be issued in logger.
    altitude_bounds: tuple = (-500.0, 40000.0)

    #: Minimum and maximum authorized mach values. If computed Mach gets beyond these limits,
    #: computation will be interrupted and a warning message will be issued in logger.
    mach_bounds: tuple = (0.0, 5.0)

    #: If True, computation will be interrupted if a parameter stops getting closer to target
    #: between two iterations (which can mean the provided thrust rate is not adapted).
    interrupt_if_getting_further_from_target: bool = True

    #: The EngineSetting value associated to the segment. Can be used in the
    #: propulsion model.
    engine_setting: EngineSetting = EngineSetting.CLIMB

    @abstractmethod
    def get_distance_to_target(
        self, flight_points: List[FlightPoint], target: FlightPoint
    ) -> float:
        """
        Computes a "distance" from last flight point to target.

        Computed does not need to have a real meaning.
        The important point is that it must be signed so that algorithm knows on
        which "side" of the target we are.
        And of course, it should be 0. if flight point is on target.

        :param flight_points: list of all currently computed flight_points
        :param target: segment target (will not contain relative values)
        :return: O. if target is attained, a non-null value otherwise
        """

    @abstractmethod
    def compute_propulsion(self, flight_point: FlightPoint):
        """
        Computes propulsion data.

        Provided flight point is modified in place.

        Generally, this method should end with::

            self.propulsion.compute_flight_points(flight_point)

        :param flight_point:
        """

    @abstractmethod
    def get_gamma_and_acceleration(self, flight_point: FlightPoint) -> Tuple[float, float]:
        """
        Computes slope angle (gamma) and acceleration.

        :param flight_point: parameters after propulsion model has been called
                             (i.e. mass, thrust and drag are available)
        :return: slope angle in radians and acceleration in m**2/s
        """

    def complete_flight_point(self, flight_point: FlightPoint):
        super().complete_flight_point(flight_point)
        if flight_point.altitude is not None:
            atm = self._get_atmosphere_point(flight_point.altitude)
            reference_force = (
                0.5 * atm.density * flight_point.true_airspeed ** 2 * self.reference_area
            )

            if self.polar:
                flight_point.CL = flight_point.mass * g / reference_force
                flight_point.CD = self.polar.cd(flight_point.CL)
            else:
                flight_point.CL = flight_point.CD = 0.0
            flight_point.drag = flight_point.CD * reference_force
        flight_point.engine_setting = self.engine_setting
        self.compute_propulsion(flight_point)
        flight_point.slope_angle, flight_point.acceleration = self.get_gamma_and_acceleration(
            flight_point
        )

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        flight_points = [start]
        previous_point_to_target = self.get_distance_to_target(flight_points, target)
        tol = 1.0e-5  # Such accuracy is not needed, but ensures reproducibility of results.
        while np.abs(previous_point_to_target) > tol:
            self._add_new_flight_point(flight_points, self.time_step)
            last_point_to_target = self.get_distance_to_target(flight_points, target)

            if last_point_to_target * previous_point_to_target < 0.0:

                # Target has been exceeded. Let's look for the exact time step using root_scalar.
                def replace_last_point(time_step):
                    """
                    Replaces last point of flight_points.

                    :param time_step: time step for new point
                    :return: new distance to target
                    """

                    if isinstance(time_step, np.ndarray):
                        # root_scalar() will provide time_step ad (1,) array, resulting
                        # in all parameters of the new flight point being also (1,) arrays.
                        # We want to avoid that
                        time_step = time_step.item()
                    del flight_points[-1]
                    self._add_new_flight_point(flight_points, time_step)
                    return self.get_distance_to_target(flight_points, target)

                root_scalar(
                    replace_last_point, x0=self.time_step, x1=self.time_step / 2.0, rtol=tol
                )
                last_point_to_target = self.get_distance_to_target(flight_points, target)
            elif (
                np.abs(last_point_to_target) > np.abs(previous_point_to_target)
                # If self.target.CL is defined, it means that we look for an optimal altitude and
                # that target altitude can move, so it would be normal to get further from target.
                and self.interrupt_if_getting_further_from_target
            ):
                # We get further from target. Let's stop without this point.
                _LOGGER.warning(
                    'Target cannot be reached in "%s". Segment computation interrupted.'
                    "Please review the segment settings, especially thrust_rate.",
                    self.name,
                )
                del flight_points[-1]
                break

            msg = self._check_values(flight_points[-1])
            if msg:
                _LOGGER.warning(msg + ' Segment computation interrupted in "%s".', self.name)
                break

            previous_point_to_target = last_point_to_target

        flight_points_df = pd.DataFrame(flight_points)
        return flight_points_df

    def compute_next_flight_point(
        self, flight_points: List[FlightPoint], time_step: float
    ) -> FlightPoint:
        """
        Computes time, altitude, speed, mass and ground distance of next flight point.

        :param flight_points: previous flight points
        :param time_step: time step for computing next point
        :return: the computed next flight point
        """
        start = flight_points[0]
        previous = flight_points[-1]
        next_point = FlightPoint()

        next_point.mass = previous.mass - self.propulsion.get_consumed_mass(previous, time_step)
        next_point.time = previous.time + time_step
        next_point.ground_distance = (
            previous.ground_distance
            + previous.true_airspeed * time_step * np.cos(previous.slope_angle)
        )
        self._compute_next_altitude(next_point, previous)

        if self.target.true_airspeed == self.CONSTANT_VALUE:
            next_point.true_airspeed = previous.true_airspeed
        elif self.target.equivalent_airspeed == self.CONSTANT_VALUE:
            next_point.equivalent_airspeed = start.equivalent_airspeed
        elif self.target.mach == self.CONSTANT_VALUE:
            next_point.mach = start.mach
        else:
            next_point.true_airspeed = previous.true_airspeed + time_step * previous.acceleration

        # The naming is not done in complete_flight_point for not naming the start point
        next_point.name = self.name
        return next_point

    def _check_values(self, flight_point: FlightPoint) -> str:
        """
        Checks that computed values are consistent.

        May be overloaded for doing specific additional checks at each time step.

        :param flight_point:
        :return: None if Ok, or an error message otherwise
        """

        if not self.mach_bounds[0] <= flight_point.mach <= self.mach_bounds[1]:
            return f"true_airspeed value {flight_point.true_airspeed:.1f}m/s is out of bound."
        if not self.altitude_bounds[0] <= flight_point.altitude <= self.altitude_bounds[1]:
            return f"Altitude value {flight_point.altitude:.0f}m is out of bound."
        if flight_point.mass <= 0.0:
            return "Negative mass value."

    def _add_new_flight_point(self, flight_points: List[FlightPoint], time_step):
        """
        Appends a new flight point to provided flight point list.

        :param flight_points: list of previous flight points, modified in place.
        :param time_step: time step for new computed flight point.
        """
        new_point = self.compute_next_flight_point(flight_points, time_step)
        self.complete_flight_point(new_point)
        flight_points.append(new_point)

    @staticmethod
    def _compute_next_altitude(next_point: FlightPoint, previous_point: FlightPoint):
        time_step = next_point.time - previous_point.time
        next_point.altitude = (
            previous_point.altitude
            + time_step * previous_point.true_airspeed * np.sin(previous_point.slope_angle)
        )

    def _get_optimal_altitude(
        self, mass: float, mach: float, altitude_guess: float = None
    ) -> float:
        """
        Computes optimal altitude for provided mass and Mach number.

        :param mass:
        :param mach:
        :return: altitude that matches optimal CL
        """

        if altitude_guess is None:
            altitude_guess = 10000.0

        def distance_to_optimum(altitude):
            atm = self._get_atmosphere_point(altitude)
            true_airspeed = mach * atm.speed_of_sound
            optimal_air_density = (
                2.0 * mass * g / (self.reference_area * true_airspeed ** 2 * self.polar.optimal_cl)
            )
            return (atm.density - optimal_air_density) * 100.0

        optimal_altitude = root_scalar(
            distance_to_optimum, x0=altitude_guess, x1=altitude_guess - 1000.0
        ).root

        return optimal_altitude


@deprecated(
    "Class FlightSegment will be removed in version 2.0. "
    "It is replaced by class AbstractTimeStepFlightSegment."
)
@dataclass
class FlightSegment(AbstractTimeStepFlightSegment, ABC):
    """
    Base class for time step computation flight segments.

    This class implements the time computation. For this computation to work, subclasses must
    implement abstract methods :meth:`get_get_distance_to_target`,
    :meth:`get_gamma_and_acceleration` and :meth:`compute_propulsion`.
    """


@dataclass
class AbstractManualThrustSegment(AbstractTimeStepFlightSegment, ABC):
    """
    Base class for computing flight segment where thrust rate is imposed.

    :ivar thrust_rate: used thrust rate. Can be set at instantiation using a keyword argument.
    """

    thrust_rate: float = 1.0

    def compute_propulsion(self, flight_point: FlightPoint):
        flight_point.thrust_rate = self.thrust_rate
        flight_point.thrust_is_regulated = False
        self.propulsion.compute_flight_points(flight_point)


@dataclass
class AbstractRegulatedThrustSegment(AbstractTimeStepFlightSegment, ABC):
    """
    Base class for computing flight segment where thrust rate is adjusted on drag.
    """

    time_step: float = 60.0

    def __post_init__(self):
        super().__post_init__()
        self.target.mach = self.CONSTANT_VALUE

    def compute_propulsion(self, flight_point: FlightPoint):
        flight_point.thrust = flight_point.drag
        flight_point.thrust_is_regulated = True
        self.propulsion.compute_flight_points(flight_point)

    def get_gamma_and_acceleration(self, flight_point: FlightPoint) -> Tuple[float, float]:
        return 0.0, 0.0


@dataclass
class AbstractFixedDurationSegment(AbstractTimeStepFlightSegment, ABC):
    """
    Base class for computing a fixed-duration segment.
    """

    time_step: float = 60.0

    def get_distance_to_target(
        self, flight_points: List[FlightPoint], target: FlightPoint
    ) -> float:
        current = flight_points[-1]
        return target.time - current.time



"""Class for very simple transition in some flight phases."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from copy import deepcopy
from dataclasses import dataclass

import pandas as pd

from fastoad.model_base import FlightPoint
from fastoad.models.performances.mission.segments.base import (
    AbstractFlightSegment,
)


@dataclass
class DummyTransitionSegment(AbstractFlightSegment, mission_file_keyword="transition"):
    """
    Computes a transient flight part in a very quick and dummy way.

    :meth:`compute_from` will return only 2 or 3 flight points.

    The second flight point is the end of transition. Its parameters are equal to those provided
    in :attr:`~fastoad.models.performances.mission.segments.base.FlightSegment.target`.

    There is an exception if target does not specify any mass (i.e. self.target.mass == 0). Then
    the mass of the second flight point is the start mass multiplied by :attr:`mass_ratio`.

    If :attr:`reserve_mass_ratio` is non-zero, a third flight point is added, with parameters equal
    to flight_point(2), except for mass where:
        mass(2) - reserve_mass_ratio * mass(3) = mass(3).
    In different words, mass(3) would be the Zero Fuel Weight (ZFW) and reserve can be
    expressed as a percentage of ZFW.
    """

    #: The ratio (aircraft mass at END of segment)/(aircraft mass at START of segment)
    mass_ratio: float = 1.0

    #: The ratio (fuel mass)/(aircraft mass at END of segment) that will be consumed at end
    #: of segment.
    reserve_mass_ratio: float = 0.0

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        end = deepcopy(target)
        end.name = self.name

        if end.mass is None:
            end.mass = start.mass * self.mass_ratio

        self.complete_flight_point_from(end, start)
        self.complete_flight_point(end)

        flight_points = [start, end]

        if self.reserve_mass_ratio > 0.0:
            reserve = deepcopy(end)
            reserve.mass = end.mass / (1.0 + self.reserve_mass_ratio)
            flight_points.append(reserve)

        return pd.DataFrame(flight_points)



"""Classes for Taxi sequences."""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from typing import Tuple

import pandas as pd

from fastoad.model_base import FlightPoint
from fastoad.models.performances.mission.segments.base import (
    AbstractFixedDurationSegment,
)
from .base import AbstractManualThrustSegment
from ..polar import Polar


@dataclass
class TaxiSegment(
    AbstractManualThrustSegment,
    AbstractFixedDurationSegment,
    mission_file_keyword="taxi",
    attribute_units=dict(true_airspeed="m/s"),
):
    """
    Class for computing Taxi phases.

    Taxi phase has a target duration (target.time should be provided) and is at
    constant altitude, speed and thrust rate.
    """

    polar: Polar = None
    reference_area: float = 1.0
    time_step: float = 60.0
    true_airspeed: float = 0.0

    def get_gamma_and_acceleration(self, flight_point: FlightPoint) -> Tuple[float, float]:
        return 0.0, 0.0

    def compute_from_start_to_target(self, start: FlightPoint, target: FlightPoint) -> pd.DataFrame:
        start.mach = None
        start.equivalent_airspeed = None
        start.true_airspeed = self.true_airspeed
        self.complete_flight_point(start)

        return super().compute_from_start_to_target(start, target)



"""
Defines the analysis and plotting functions for postprocessing
"""
#  This file is part of FAST-OAD : A framework for rapid Overall Aircraft Design
#  Copyright (C) 2022 ONERA & ISAE-SUPAERO
#  FAST is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from openmdao.utils.units import convert_units
from plotly.subplots import make_subplots

from fastoad.io import VariableIO
from fastoad.openmdao.variables import VariableList

COLS = px.colors.qualitative.Plotly


# pylint: disable-msg=too-many-locals
def wing_geometry_plot(
    aircraft_file_path: str, name=None, fig=None, *, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the top view of the wing.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided,
                           default format will be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    wing_kink_leading_edge_x = variables["data:geometry:wing:kink:leading_edge:x:local"].value[0]
    wing_tip_leading_edge_x = variables["data:geometry:wing:tip:leading_edge:x:local"].value[0]
    wing_root_y = variables["data:geometry:wing:root:y"].value[0]
    wing_kink_y = variables["data:geometry:wing:kink:y"].value[0]
    wing_tip_y = variables["data:geometry:wing:tip:y"].value[0]
    wing_root_chord = variables["data:geometry:wing:root:chord"].value[0]
    wing_kink_chord = variables["data:geometry:wing:kink:chord"].value[0]
    wing_tip_chord = variables["data:geometry:wing:tip:chord"].value[0]

    mean_aerodynamic_chord = variables["data:geometry:wing:MAC:length"].value[0]
    mac25_x_position = variables["data:geometry:wing:MAC:at25percent:x"].value[0]
    distance_root_mac_chords = variables["data:geometry:wing:MAC:leading_edge:x:local"].value[0]
    # pylint: disable=invalid-name # that's a common naming
    y = np.array(
        [0, wing_root_y, wing_kink_y, wing_tip_y, wing_tip_y, wing_kink_y, wing_root_y, 0, 0]
    )
    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((-y, y))

    # pylint: disable=invalid-name # that's a common naming
    x = np.array(
        [
            0,
            0,
            wing_kink_leading_edge_x,
            wing_tip_leading_edge_x,
            wing_tip_leading_edge_x + wing_tip_chord,
            wing_kink_leading_edge_x + wing_kink_chord,
            wing_root_chord,
            wing_root_chord,
            0,
        ]
    )

    x = x + mac25_x_position - 0.25 * mean_aerodynamic_chord - distance_root_mac_chords
    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x, x))

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=y, y=x, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig.layout = go.Layout(yaxis=dict(scaleanchor="x", scaleratio=1))

    fig = go.FigureWidget(fig)

    fig.update_layout(title_text="Wing Geometry", title_x=0.5, xaxis_title="y", yaxis_title="x")

    return fig


# pylint: disable-msg=too-many-locals
def aircraft_geometry_plot(
    aircraft_file_path: str, name=None, fig=None, *, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the top view of the wing.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided,
                           default format will be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    # Wing parameters
    wing_kink_leading_edge_x = variables["data:geometry:wing:kink:leading_edge:x:local"].value[0]
    wing_tip_leading_edge_x = variables["data:geometry:wing:tip:leading_edge:x:local"].value[0]
    wing_root_y = variables["data:geometry:wing:root:y"].value[0]
    wing_kink_y = variables["data:geometry:wing:kink:y"].value[0]
    wing_tip_y = variables["data:geometry:wing:tip:y"].value[0]
    wing_root_chord = variables["data:geometry:wing:root:chord"].value[0]
    wing_kink_chord = variables["data:geometry:wing:kink:chord"].value[0]
    wing_tip_chord = variables["data:geometry:wing:tip:chord"].value[0]

    y_wing = np.array(
        [0, wing_root_y, wing_kink_y, wing_tip_y, wing_tip_y, wing_kink_y, wing_root_y, 0, 0]
    )

    x_wing = np.array(
        [
            0,
            0,
            wing_kink_leading_edge_x,
            wing_tip_leading_edge_x,
            wing_tip_leading_edge_x + wing_tip_chord,
            wing_kink_leading_edge_x + wing_kink_chord,
            wing_root_chord,
            wing_root_chord,
            0,
        ]
    )

    # Horizontal Tail parameters
    ht_root_chord = variables["data:geometry:horizontal_tail:root:chord"].value[0]
    ht_tip_chord = variables["data:geometry:horizontal_tail:tip:chord"].value[0]
    ht_span = variables["data:geometry:horizontal_tail:span"].value[0]
    ht_sweep_0 = variables["data:geometry:horizontal_tail:sweep_0"].value[0]

    ht_tip_leading_edge_x = ht_span / 2.0 * np.tan(ht_sweep_0 * np.pi / 180.0)

    y_ht = np.array([0, ht_span / 2.0, ht_span / 2.0, 0.0, 0.0])

    x_ht = np.array(
        [0, ht_tip_leading_edge_x, ht_tip_leading_edge_x + ht_tip_chord, ht_root_chord, 0]
    )

    # Fuselage parameters
    fuselage_max_width = variables["data:geometry:fuselage:maximum_width"].value[0]
    fuselage_length = variables["data:geometry:fuselage:length"].value[0]
    fuselage_front_length = variables["data:geometry:fuselage:front_length"].value[0]
    fuselage_rear_length = variables["data:geometry:fuselage:rear_length"].value[0]

    x_fuselage = np.array(
        [
            0.0,
            0.0,
            fuselage_front_length,
            fuselage_length - fuselage_rear_length,
            fuselage_length,
            fuselage_length,
        ]
    )

    y_fuselage = np.array(
        [
            0.0,
            fuselage_max_width / 4.0,
            fuselage_max_width / 2.0,
            fuselage_max_width / 2.0,
            fuselage_max_width / 4.0,
            0.0,
        ]
    )

    # CGs
    wing_25mac_x = variables["data:geometry:wing:MAC:at25percent:x"].value[0]
    wing_mac_length = variables["data:geometry:wing:MAC:length"].value[0]
    local_wing_mac_le_x = variables["data:geometry:wing:MAC:leading_edge:x:local"].value[0]
    local_ht_25mac_x = variables["data:geometry:horizontal_tail:MAC:at25percent:x:local"].value[0]
    ht_distance_from_wing = variables[
        "data:geometry:horizontal_tail:MAC:at25percent:x:from_wingMAC25"
    ].value[0]

    x_wing = x_wing + wing_25mac_x - 0.25 * wing_mac_length - local_wing_mac_le_x
    x_ht = x_ht + wing_25mac_x + ht_distance_from_wing - local_ht_25mac_x

    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x_fuselage, x_wing, x_ht))
    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((y_fuselage, y_wing, y_ht))

    # pylint: disable=invalid-name # that's a common naming
    y = np.concatenate((-y, y))
    # pylint: disable=invalid-name # that's a common naming
    x = np.concatenate((x, x))

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=y, y=x, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig.layout = go.Layout(yaxis=dict(scaleanchor="x", scaleratio=1))

    fig = go.FigureWidget(fig)

    fig.update_layout(title_text="Aircraft Geometry", title_x=0.5, xaxis_title="y", yaxis_title="x")

    return fig


def drag_polar_plot(
    aircraft_file_path: str, name=None, fig=None, *, file_formatter=None
) -> go.FigureWidget:
    """
    Returns a figure plot of the aircraft drag polar.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided,
                           default format will be assumed.
    :return: wing plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    # pylint: disable=invalid-name # that's a common naming
    cd = np.asarray(variables["data:aerodynamics:aircraft:cruise:CD"].value)
    # pylint: disable=invalid-name # that's a common naming
    cl = np.asarray(variables["data:aerodynamics:aircraft:cruise:CL"].value)

    # TODO: remove filtering one models provide proper bounds
    cd_short = cd[cd <= 2.0]
    cl_short = cl[cd <= 2.0]

    if fig is None:
        fig = go.Figure()

    scatter = go.Scatter(x=cd_short, y=cl_short, mode="lines+markers", name=name)

    fig.add_trace(scatter)

    fig = go.FigureWidget(fig)

    fig.update_layout(title_text="Drag Polar", title_x=0.5, xaxis_title="Cd", yaxis_title="Cl")

    return fig


def mass_breakdown_bar_plot(
    aircraft_file_path: str,
    name=None,
    fig=None,
    *,
    file_formatter=None,
    input_mass_name="data:weight:aircraft:MTOW",
) -> go.FigureWidget:
    """
    Returns a figure plot of the aircraft mass breakdown using bar plots.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param name: name to give to the trace added to the figure
    :param fig: existing figure to which add the plot
    :param file_formatter: the formatter that defines the format of data file. If not provided,
                           default format will be assumed.
    :param input_mass_name: the variable name for the mass input as defined in the mission
                            definition file.
    :return: bar plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    var_names_and_new_units = {
        input_mass_name: "kg",
        "data:weight:aircraft:OWE": "kg",
        "data:weight:aircraft:payload": "kg",
        "data:weight:aircraft:sizing_onboard_fuel_at_input_weight": "kg",
    }

    # pylint: disable=unbalanced-tuple-unpacking # It is balanced for the parameters provided
    mtow, owe, payload, fuel_mission = _get_variable_values_with_new_units(
        variables, var_names_and_new_units
    )

    if fig is None:
        fig = make_subplots(
            rows=1,
            cols=2,
            subplot_titles=("Maximum Take-Off Weight Breakdown", "Overall Weight Empty Breakdown"),
        )

    # Same color for each aircraft configuration
    i = int(len(fig.data) / 2) % 10

    weight_labels = ["MTOW", "OWE", "Fuel - Mission", "Payload"]
    weight_values = [mtow, owe, fuel_mission, payload]
    fig.add_trace(
        go.Bar(name="", x=weight_labels, y=weight_values, marker_color=COLS[i], showlegend=False),
        row=1,
        col=1,
    )

    # Get data:weight decomposition
    main_weight_values, main_weight_names, _ = _data_weight_decomposition(variables, owe=None)
    fig.add_trace(
        go.Bar(name=name, x=main_weight_names, y=main_weight_values, marker_color=COLS[i]),
        row=1,
        col=2,
    )

    fig.update_layout(yaxis_title="[kg]")

    return fig


def mass_breakdown_sun_plot(
    aircraft_file_path: str,
    *,
    file_formatter=None,
    input_mass_name="data:weight:aircraft:MTOW",
):
    """
    Returns a figure sunburst plot of the mass breakdown.
    On the left a MTOW sunburst and on the right a OWE sunburst.
    Different designs can be superposed by providing an existing fig.
    Each design can be provided a name.

    :param aircraft_file_path: path of data file
    :param file_formatter: the formatter that defines the format of data file. If not provided,
                           default format will be assumed.
    :param input_mass_name: the variable name for the mass input as defined in the mission
                            definition file.
    :return: sunburst plot figure
    """
    variables = VariableIO(aircraft_file_path, file_formatter).read()

    var_names_and_new_units = {
        input_mass_name: "kg",
        "data:weight:aircraft:OWE": "kg",
        "data:weight:aircraft:payload": "kg",
        "data:weight:aircraft:sizing_onboard_fuel_at_input_weight": "kg",
    }

    # pylint: disable=unbalanced-tuple-unpacking # It is balanced for the parameters provided
    mtow, owe, payload, onboard_fuel_at_takeoff = _get_variable_values_with_new_units(
        variables, var_names_and_new_units
    )

    # TODO: Deal with this in a more generic manner ?
    if round(mtow, 6) == round(owe + payload + onboard_fuel_at_takeoff, 6):
        mtow = owe + payload + onboard_fuel_at_takeoff

    fig = make_subplots(1, 2, specs=[[{"type": "domain"}, {"type": "domain"}]])

    fig.add_trace(
        go.Sunburst(
            labels=[
                "MTOW" + "<br>" + str(int(mtow)) + " [kg]",
                "payload"
                + "<br>"
                + str(int(payload))
                + " [kg] ("
                + str(round(payload / mtow * 100, 1))
                + "%)",
                "onboard_fuel_at_takeoff"
                + "<br>"
                + str(int(onboard_fuel_at_takeoff))
                + " [kg] ("
                + str(round(onboard_fuel_at_takeoff / mtow * 100, 1))
                + "%)",
                "OWE" + "<br>" + str(int(owe)) + " [kg] (" + str(round(owe / mtow * 100, 1)) + "%)",
            ],
            parents=[
                "",
                "MTOW" + "<br>" + str(int(mtow)) + " [kg]",
                "MTOW" + "<br>" + str(int(mtow)) + " [kg]",
                "MTOW" + "<br>" + str(int(mtow)) + " [kg]",
            ],
            values=[mtow, payload, onboard_fuel_at_takeoff, owe],
            branchvalues="total",
        ),
        1,
        1,
    )

    # Get data:weight 2-levels decomposition
    categories_values, categories_names, categories_labels = _data_weight_decomposition(
        variables, owe=owe
    )

    sub_categories_values = []
    sub_categories_names = []
    sub_categories_parent = []
    for variable in variables.names():
        name_split = variable.split(":")
        if isinstance(name_split, list) and len(name_split) >= 5:
            parent_name = name_split[2]
            if parent_name in categories_names and name_split[-1] == "mass":
                variable_name = "_".join(name_split[3:-1])
                sub_categories_values.append(
                    convert_units(variables[variable].value[0], variables[variable].units, "kg")
                )
                sub_categories_parent.append(categories_labels[categories_names.index(parent_name)])
                sub_categories_names.append(variable_name)

    # Define figure data
    figure_labels = ["OWE" + "<br>" + str(int(owe)) + " [kg]"]
    figure_labels.extend(categories_labels)
    figure_labels.extend(sub_categories_names)
    figure_parents = [""]
    for _ in categories_names:
        figure_parents.append("OWE" + "<br>" + str(int(owe)) + " [kg]")
    figure_parents.extend(sub_categories_parent)
    figure_values = [owe]
    figure_values.extend(categories_values)
    figure_values.extend(sub_categories_values)

    # Plot figure
    fig.add_trace(
        go.Sunburst(
            labels=figure_labels, parents=figure_parents, values=figure_values, branchvalues="total"
        ),
        1,
        2,
    )

    fig.update_layout(title_text="Mass Breakdown", title_x=0.5)

    return fig


def _get_variable_values_with_new_units(
    variables: VariableList, var_names_and_new_units: Dict[str, str]
):
    """
    Returns the value of the requested variable names with respect to their new units in the order
    in which their were given. This function works only for variable of value with shape=1 or float.

    :param variables: instance containing variables information
    :param var_names_and_new_units: dictionary of the variable names as keys and units as value
    :return: values of the requested variables with respect to their new units
    """
    new_values = []
    for variable_name, unit in var_names_and_new_units.items():
        new_values.append(
            convert_units(variables[variable_name].value[0], variables[variable_name].units, unit)
        )

    return new_values


def _data_weight_decomposition(variables: VariableList, owe=None):
    """
    Returns the two level weight decomposition of MTOW and optionally the decomposition of owe
    subcategories.

    :param variables: instance containing variables information
    :param owe: value of OWE, if provided names of owe subcategories will be provided
    :return: variable values, names and optionally owe subcategories names
    """
    category_values = []
    category_names = []
    owe_subcategory_names = []
    for variable in variables.names():
        name_split = variable.split(":")
        if isinstance(name_split, list) and len(name_split) == 4:
            if name_split[0] + name_split[1] + name_split[3] == "dataweightmass" and not (
                "aircraft" in name_split[2]
            ):
                category_values.append(
                    convert_units(variables[variable].value[0], variables[variable].units, "kg")
                )
                category_names.append(name_split[2])
                if owe:
                    owe_subcategory_names.append(
                        name_split[2]
                        + "<br>"
                        + str(int(variables[variable].value[0]))
                        + " [kg] ("
                        + str(round(variables[variable].value[0] / owe * 100, 1))
                        + "%)"
                    )
    if owe:
        result = category_values, category_names, owe_subcategory_names
    else:
        result = category_values, category_names, None

    return result



def compute_wing_area(mtow, wing_loading):
    """
    Computes the wing area based on the provided MTOW and wing loading

    :param mtow: Max Take-Off Weight, in kg
    :param wing_loading: Wing loading, in kg/m2

    :return wing_area: Wing area, in m2
    """

    # Computation of the wing area
    wing_area = mtow / wing_loading

    return wing_area



from IPython.display import display

import ipywidgets as widgets


class ChoiceForOWE:
    def __init__(self):

        box_layout = widgets.Layout(
            display="flex", flex_flow="column", align_items="stretch", width="100%"
        )

        self.choice_button = widgets.ToggleButtons(
            options=["Corrected component", "Exercise component"],
            disabled=False,
            button_style="info",
            tooltips=[
                "Toggle this button to use the correct component",
                "Toggle this button to use the component left as exercise",
            ],
            icon=["fa-check-square-o", "fa-wrench"],
            style={"button_width": "48%"},
            layout=box_layout,
        )
        self.choice = "correct"

        self.choice_button.observe(self.select_component)

        display(self.choice_button)

    def select_component(self, b):

        if self.choice_button.value == "Corrected component":
            self.choice = "correct"
        else:
            self.choice = "exercise"



from ..pure_python.geometry.geometry import compute_geometry
from ..pure_python.aerodynamics.aerodynamics import compute_aerodynamics
from ..pure_python.mass.mass import compute_mass
from ..pure_python.performance.performance import compute_performance
from ..pure_python.update_mtow.update_mtow import update_mtow


def mtow_loop_scipy(x, wing_loading, cruise_altitude, cruise_speed, mission_range, payload, tsfc):
    """
    Gather all the module main functions in the program main function that will compute a new
    MTOW based on an old MTOW

    :param x: Tuple containing the Old Max Take-Off Weight, in kg and the aspect ration with no unit
    :param wing_loading: Wing loading, in kg/m2
    :param cruise_speed: Cruise speed, in m/s
    :param cruise_altitude: Cruise altitude, in m
    :param payload: the payload mass, in kg
    :param mission_range: the mission range, in m
    :param tsfc: the thrust specific fuel consumption, in kg/N/s

    :return updated_mtow: Max Take-Off Weight computed based on the old value
    """

    # Let's unpack the tuple containing the two values that will be used for the optimization
    mtow = x[0]
    aspect_ratio = x[1]

    # Let's start by computing the aircraft geometry
    wing_area = compute_geometry(mtow=mtow, wing_loading=wing_loading)

    # Let's now compute its aerodynamic properties
    l_d = compute_aerodynamics(
        cruise_altitude=cruise_altitude,
        cruise_speed=cruise_speed,
        mtow=mtow,
        wing_area=wing_area,
        aspect_ratio=aspect_ratio,
    )

    # We can now compute its structural mass
    owe = compute_mass(mtow=mtow, wing_area=wing_area, aspect_ratio=aspect_ratio)

    # Let's now get the mission performances
    mission_fuel = compute_performance(
        owe=owe,
        payload=payload,
        mission_range=mission_range,
        tsfc=tsfc,
        l_d=l_d,
        cruise_speed=cruise_speed,
    )

    # Finally, let's compute the new mtow
    updated_mtow = update_mtow(
        owe=owe,
        payload=payload,
        mission_fuel=mission_fuel,
    )

    return updated_mtow
